# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Dict, Optional

from fairseq.incremental_decoding_utils import with_incremental_state
from fairseq.models import FairseqDecoder
from torch import Tensor


logger = logging.getLogger(__name__)


@with_incremental_state
class FairseqIncrementalDecoder(FairseqDecoder):
    """Base class for incremental decoders.

    Incremental decoding is a special mode at inference time where the Model
    only receives a single timestep of input corresponding to the previous
    output token (for teacher forcing) and must produce the next output
    *incrementally*. Thus the model must cache any long-term state that is
    needed about the sequence, e.g., hidden states, convolutional states, etc.

    Compared to the standard :class:`FairseqDecoder` interface, the incremental
    decoder interface allows :func:`forward` functions to take an extra keyword
    argument (*incremental_state*) that can be used to cache state across
    time-steps.

    The :class:`FairseqIncrementalDecoder` interface also defines the
    :func:`reorder_incremental_state` method, which is used during beam search
    to select and reorder the incremental state based on the selection of beams.

    To learn more about how incremental decoding works, refer to `this blog
    <http://www.telesens.co/2019/04/21/understanding-incremental-decoding-in-fairseq/>`_.
    """

    def __init__(self, dictionary):
        super().__init__(dictionary)

    def forward(
        self, prev_output_tokens, encoder_out=None, incremental_state=None, **kwargs
    ):
        """
        Args:
            prev_output_tokens (LongTensor): shifted output tokens of shape
                `(batch, tgt_len)`, for teacher forcing
            encoder_out (dict, optional): output from the encoder, used for
                encoder-side attention
            incremental_state (dict, optional): dictionary used for storing
                state during :ref:`Incremental decoding`

        Returns:
            tuple:
                - the decoder's output of shape `(batch, tgt_len, vocab)`
                - a dictionary with any model-specific outputs
        """
        raise NotImplementedError

    def extract_features(
        self, prev_output_tokens, encoder_out=None, incremental_state=None, **kwargs
    ):
        """
        Returns:
            tuple:
                - the decoder's features of shape `(batch, tgt_len, embed_dim)`
                - a dictionary with any model-specific outputs
        """
        raise NotImplementedError

    def reorder_incremental_state(
        self,
        incremental_state: Dict[str, Dict[str, Optional[Tensor]]],
        new_order: Tensor,
    ):
        """Reorder incremental state.

        This will be called when the order of the input has changed from the
        previous time step. A typical use case is beam search, where the input
        order changes between time steps based on the selection of beams.
        """
        pass

    def reorder_incremental_state_scripting(
        self,
        incremental_state: Dict[str, Dict[str, Optional[Tensor]]],
        new_order: Tensor,
    ):
        """Main entry point for reordering the incremental state.

        Due to limitations in TorchScript, we call this function in
        :class:`fairseq.sequence_generator.SequenceGenerator` instead of
        calling :func:`reorder_incremental_state` directly.
        """
        for module in self.modules():
            if hasattr(module, "reorder_incremental_state"):
                result = module.reorder_incremental_state(incremental_state, new_order)
                if result is not None:
                    incremental_state = result

    def set_beam_size(self, beam_size):
        """Sets the beam size in the decoder and all children."""
        if getattr(self, "_beam_size", -1) != beam_size:
            seen = set()

            def apply_set_beam_size(module):
                if (
                    module != self
                    and hasattr(module, "set_beam_size")
                    and module not in seen
                ):
                    seen.add(module)
                    module.set_beam_size(beam_size)

            self.apply(apply_set_beam_size)
            self._beam_size = beam_size



import io
import requests
import os
import subprocess
from threading import Thread
from time import perf_counter


import modal


app = modal.App("phi4_multimodal")
img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "clang", "cmake", "ninja-build")
    .pip_install(
        "transformers==4.48.2",
        "torch==2.6.0",
        "torchaudio==2.6.0",
        "torchvision==0.21.0",
    )
    .pip_install("wheel")
    # https://github.com/Dao-AILab/flash-attention/releases/tag/v2.7.4.post1
    .pip_install("flash-attn", extra_options="--no-build-isolation")
    .pip_install("soundfile", "accelerate", "pillow", "scipy", "backoff", "peft")
    .pip_install("qwen_omni_utils")
    .env(
        {
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
            "TQDM_DISABLE": "1",
            "LLM_MODEL": os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct"),
        }
    )
)

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)


with img.imports():
    import torch
    from PIL import Image
    import soundfile as sf
    from transformers import AutoModelForCausalLM, AutoProcessor, GenerationConfig
    from transformers.generation.streamers import TextIteratorStreamer
    from qwen_omni_utils import process_mm_info

    # Define prompt structure
    USER_PROMPT = "<|user|>"
    ASSISTANT_PROMPT = "<|assistant|>"
    PROMPT_SUFFIX = "<|end|>"


@app.function(
    gpu=os.getenv("IMAGE_GPU", None),
    cpu=2.0,
    retries=1,
    image=img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        ASSETS_DIR: assets_dir,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
def run(func):
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    gpu_prop = torch.cuda.get_device_properties("cuda")
    print(gpu_prop)

    func(gpu_prop)


def print_model_params(model: torch.nn.Module, extra_info="", f=None):
    # print the number of parameters in the model
    model_million_params = sum(p.numel() for p in model.parameters()) / 1e6
    print(model, file=f)
    print(f"{extra_info} {model_million_params} M parameters", file=f)


def dump_model(gpu_prop):
    for model_name in [
        "microsoft/Phi-4-mini-instruct",
        "microsoft/Phi-4-multimodal-instruct",
    ]:
        model_path = os.path.join(HF_MODEL_DIR, model_name)
        processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)

        model = AutoModelForCausalLM.from_pretrained(
            model_path,
            torch_dtype=torch.bfloat16,
            trust_remote_code=True,
            attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        ).to("cuda")

        model = model.eval()
        print(f"{model.config=}")
        print("model.config._attn_implementation:", model.config._attn_implementation)

        file_path = os.path.join(model_path, "model.txt")
        with open(file_path, "w") as f:
            print(f"{processor=}", file=f)
            print_model_params(model, f"{model_name}", f)

        # Load generation config
        generation_config = GenerationConfig.from_pretrained(model_path)
        print(f"{generation_config=}")

        del model
        torch.cuda.empty_cache()


def text2text(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    prompt = (
        f"{USER_PROMPT}what is the answer for 1+1? Explain it.{PROMPT_SUFFIX}{ASSISTANT_PROMPT}"
    )
    inputs = processor(prompt, images=None, return_tensors="pt").to("cuda")
    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generate_ids = model.generate(
        **inputs,
        max_new_tokens=1000,
        generation_config=generation_config,
    )
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]
    response = processor.batch_decode(
        generate_ids,
        # skip_special_tokens=True,
        clean_up_tokenization_spaces=False,
    )[0]
    print(f">>> Response\n{response}")


def text2text_stream(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    prompt = (
        f"{USER_PROMPT}what is the answer for 1+1? Explain it.{PROMPT_SUFFIX}{ASSISTANT_PROMPT}"
    )
    inputs = processor(prompt, images=None, return_tensors="pt").to("cuda")
    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    for i in range(2):
        streamer = TextIteratorStreamer(
            tokenizer=processor, skip_prompt=True, skip_special_tokens=True
        )
        generation_kwargs = dict(
            **inputs,
            # do_sample=False,
            do_sample=True,
            temperature=0.9,
            top_k=10,
            top_p=0.9,
            num_beams=1,
            repetition_penalty=1.1,
            max_new_tokens=256,
            use_cache=True,
            streamer=streamer,
        )
        thread = Thread(target=model.generate, kwargs=generation_kwargs)
        thread.start()
        generated_text = ""
        start = perf_counter()
        times = []
        with torch.inference_mode():
            for new_text in streamer:
                times.append(perf_counter() - start)
                print(new_text, end="", flush=True)
                generated_text += new_text
                start = perf_counter()
        print(f"\n{i}. {generated_text=} TTFT: {times[0]:.2f}s total time: {sum(times):.2f}s")


def single_vision2text(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    # single-image prompt
    prompt = f"{USER_PROMPT}<|image_1|>描述图片中的内容?{PROMPT_SUFFIX}{ASSISTANT_PROMPT}"
    url = "https://www.ilankelman.org/stopsigns/australia.jpg"
    image = Image.open(requests.get(url, stream=True).raw)
    inputs = processor(text=prompt, images=image, return_tensors="pt").to("cuda")
    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generate_ids = model.generate(
        **inputs,
        max_new_tokens=1000,
        generation_config=generation_config,
    )
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]
    response = processor.batch_decode(
        generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
    )[0]
    print(f">>> Response\n{response}")


def multi_vision2text(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    # multi-image prompt
    images = []
    placeholder = ""
    for i in range(1, 5):
        url = f"https://image.slidesharecdn.com/azureintroduction-191206101932/75/Introduction-to-Microsoft-Azure-Cloud-{i}-2048.jpg"
        images.append(Image.open(requests.get(url, stream=True).raw))
        placeholder += f"<|image_{i}|>"
    messages = [
        {"role": "user", "content": placeholder + "Summarize the deck of slides."},
    ]

    prompt = processor.tokenizer.apply_chat_template(
        messages, tokenize=False, add_generation_prompt=True
    )
    print(f">>> Prompt\n{prompt}")

    inputs = processor(text=prompt, images=images, return_tensors="pt").to("cuda")
    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generate_ids = model.generate(
        **inputs,
        max_new_tokens=1000,
        generation_config=generation_config,
    )
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]
    response = processor.batch_decode(
        generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
    )[0]
    print(f">>> Response\n{response}")


def single_turn_vision2text(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    # chat template
    chat = [
        {"role": "system", "content": f"请用中文回复"},
        {"role": "user", "content": f"<|image_1|>请描述图片中的内容?"},
    ]
    prompt = processor.tokenizer.apply_chat_template(
        chat, tokenize=False, add_generation_prompt=True
    )
    # need to remove last <|endoftext|> if it is there, which is used for training, not inference. For training, make sure to add <|endoftext|> in the end.
    if prompt.endswith("<|endoftext|>"):
        prompt = prompt.rstrip("<|endoftext|>")
    print(f">>> Prompt\n{prompt}")

    url = "https://www.ilankelman.org/stopsigns/australia.jpg"
    image = Image.open(requests.get(url, stream=True).raw)
    inputs = processor(text=prompt, images=image, return_tensors="pt").to("cuda")
    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generate_ids = model.generate(
        **inputs,
        max_new_tokens=1000,
        generation_config=generation_config,
    )
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]
    response = processor.batch_decode(
        generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
    )[0]
    print(f">>> Response\n{response}")


def multi_turn_vision2text(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    # chat template
    chat = [
        {"role": "system", "content": f"请用中文回复"},
        {"role": "user", "content": f"<|image_1|>What is shown in this image?"},
        {
            "role": "assistant",
            "content": "The image depicts a street scene with a prominent red stop sign in the foreground. The background showcases a building with traditional Chinese architecture, characterized by its red roof and ornate decorations. There are also several statues of lions, which are common in Chinese culture, positioned in front of the building. The street is lined with various shops and businesses, and there's a car passing by.",
        },
        {"role": "user", "content": "<|image_2|>What is so special about this image"},
    ]
    prompt = processor.tokenizer.apply_chat_template(
        chat, tokenize=False, add_generation_prompt=True
    )
    # need to remove last <|endoftext|> if it is there, which is used for training, not inference. For training, make sure to add <|endoftext|> in the end.
    if prompt.endswith("<|endoftext|>"):
        prompt = prompt.rstrip("<|endoftext|>")

    print(f">>> Prompt\n{prompt}")

    url = "https://www.ilankelman.org/stopsigns/australia.jpg"
    image = Image.open(requests.get(url, stream=True).raw)
    inputs = processor(text=prompt, images=[image, image], return_tensors="pt").to("cuda")
    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generate_ids = model.generate(
        **inputs,
        max_new_tokens=1000,
        generation_config=generation_config,
    )
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]
    response = processor.batch_decode(
        generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
    )[0]
    print(f">>> Response\n{response}")


def speech_vision2text(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    # speech-image prompt
    AUDIO_FILE_1 = os.path.join(model_path, "examples/what_is_the_traffic_sign_in_the_image.wav")
    prompt = f"{USER_PROMPT}<|image_1|><|speech_1|>{PROMPT_SUFFIX}{ASSISTANT_PROMPT}"
    url = "https://www.ilankelman.org/stopsigns/australia.jpg"
    image = Image.open(requests.get(url, stream=True).raw)
    speech = sf.read(AUDIO_FILE_1)
    inputs = processor(text=prompt, images=[image], speechs=[speech], return_tensors="pt").to(
        "cuda"
    )

    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generate_ids = model.generate(
        **inputs,
        max_new_tokens=1000,
        generation_config=generation_config,
    )
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]
    response = processor.batch_decode(
        generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
    )[0]
    print(f">>> Response\n{response}")


def asr(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    # single speech prompt
    AUDIO_FILE_1 = os.path.join(model_path, "examples/what_is_the_traffic_sign_in_the_image.wav")
    speech_prompt = "Based on the attached audio, generate a comprehensive text transcription of the spoken content."
    prompt = f"{USER_PROMPT}<|audio_1|>{speech_prompt}{PROMPT_SUFFIX}{ASSISTANT_PROMPT}"
    audio = sf.read(AUDIO_FILE_1)
    inputs = processor(text=prompt, audios=[audio], return_tensors="pt").to("cuda")
    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generate_ids = model.generate(
        **inputs,
        max_new_tokens=1000,
        generation_config=generation_config,
    )
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]
    response = processor.batch_decode(
        generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
    )[0]
    print(f">>> Response\n{response}")


def multi_asr(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    # multi speech prompt
    AUDIO_FILE_1 = os.path.join(model_path, "examples/what_is_the_traffic_sign_in_the_image.wav")
    AUDIO_FILE_2 = os.path.join(model_path, "examples/what_is_shown_in_this_image.wav")
    audio_1 = sf.read(AUDIO_FILE_2)
    audio_2 = sf.read(AUDIO_FILE_1)
    chat = [
        {
            "role": "user",
            "content": f"<|audio_1|>Based on the attached audio, generate a comprehensive text transcription of the spoken content.",
        },
        {
            "role": "assistant",
            "content": "What is shown in this image.",
        },
        {
            "role": "user",
            "content": f"<|audio_2|>Based on the attached audio, generate a comprehensive text transcription of the spoken content.",
        },
    ]
    prompt = processor.tokenizer.apply_chat_template(
        chat, tokenize=False, add_generation_prompt=True
    )
    # need to remove last <|endoftext|> if it is there, which is used for training, not inference. For training, make sure to add <|endoftext|> in the end.
    if prompt.endswith("<|endoftext|>"):
        prompt = prompt.rstrip("<|endoftext|>")

    print(f">>> Prompt\n{prompt}")

    inputs = processor(text=prompt, audios=[audio_1, audio_2], return_tensors="pt").to("cuda")

    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generate_ids = model.generate(
        **inputs,
        max_new_tokens=1000,
        generation_config=generation_config,
    )
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]
    response = processor.batch_decode(
        generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
    )[0]
    print(f">>> Response\n{response}")


def multi_turn_vision_speech2text(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    # multi speech prompt
    AUDIO_FILE_1 = os.path.join(model_path, "examples/what_is_the_traffic_sign_in_the_image.wav")
    AUDIO_FILE_2 = os.path.join(model_path, "examples/what_is_shown_in_this_image.wav")
    audio_1 = sf.read(AUDIO_FILE_2)
    audio_2 = sf.read(AUDIO_FILE_1)
    chat = [
        {"role": "user", "content": f"<|image_1|><|audio_1|>"},
        {
            "role": "assistant",
            "content": "The image depicts a street scene with a prominent red stop sign in the foreground. The background showcases a building with traditional Chinese architecture, characterized by its red roof and ornate decorations. There are also several statues of lions, which are common in Chinese culture, positioned in front of the building. The street is lined with various shops and businesses, and there's a car passing by.",
        },
        {"role": "user", "content": f"<|audio_2|>"},
    ]
    url = "https://www.ilankelman.org/stopsigns/australia.jpg"
    image = Image.open(requests.get(url, stream=True).raw)
    prompt = processor.tokenizer.apply_chat_template(
        chat, tokenize=False, add_generation_prompt=True
    )
    # need to remove last <|endoftext|> if it is there, which is used for training, not inference. For training, make sure to add <|endoftext|> in the end.
    if prompt.endswith("<|endoftext|>"):
        prompt = prompt.rstrip("<|endoftext|>")

    print(f">>> Prompt\n{prompt}")

    inputs = processor(
        text=prompt,
        images=[image],
        # audios=[(audio_1[0], 16000), (audio_2[0], 16000)],
        audios=[audio_1, audio_2],
        return_tensors="pt",
    ).to("cuda")

    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generate_ids = model.generate(
        **inputs,
        max_new_tokens=1000,
        generation_config=generation_config,
    )
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]
    response = processor.batch_decode(
        generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
    )[0]
    print(f">>> Response\n{response}")


def multi_turn_vision_speech2text_stream(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
    ).to("cuda")
    model = model.eval()

    # Load generation config
    generation_config = GenerationConfig.from_pretrained(model_path)
    print(f"{generation_config=}")

    # multi turn speech + image prompt
    url = "https://www.ilankelman.org/stopsigns/australia.jpg"
    image = Image.open(requests.get(url, stream=True).raw)
    AUDIO_FILE_1 = os.path.join(model_path, "examples/what_is_the_traffic_sign_in_the_image.wav")
    AUDIO_FILE_2 = os.path.join(model_path, "examples/what_is_shown_in_this_image.wav")
    audio_1 = sf.read(AUDIO_FILE_2)
    audio_2 = sf.read(AUDIO_FILE_1)
    chat = [
        {"role": "user", "content": f"<|image_1|><|audio_1|>"},
        {
            "role": "assistant",
            "content": "The image depicts a street scene with a prominent red stop sign in the foreground. The background showcases a building with traditional Chinese architecture, characterized by its red roof and ornate decorations. There are also several statues of lions, which are common in Chinese culture, positioned in front of the building. The street is lined with various shops and businesses, and there's a car passing by.",
        },
        {"role": "user", "content": f"<|audio_2|>"},
    ]
    prompt = processor.tokenizer.apply_chat_template(
        chat, tokenize=False, add_generation_prompt=True
    )
    # need to remove last <|endoftext|> if it is there, which is used for training, not inference. For training, make sure to add <|endoftext|> in the end.
    if prompt.endswith("<|endoftext|>"):
        prompt = prompt.rstrip("<|endoftext|>")

    print(f">>> Prompt\n{prompt}")

    inputs = processor(
        text=prompt, images=[image], audios=[audio_1, audio_2], return_tensors="pt"
    ).to("cuda")

    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    for i in range(2):
        streamer = TextIteratorStreamer(
            tokenizer=processor, skip_prompt=True, skip_special_tokens=True
        )
        generation_kwargs = dict(
            **inputs,
            # do_sample=False,
            do_sample=True,
            temperature=0.9,
            top_k=10,
            top_p=0.9,
            num_beams=1,
            repetition_penalty=1.1,
            max_new_tokens=256,
            use_cache=True,
            streamer=streamer,
        )
        thread = Thread(target=model.generate, kwargs=generation_kwargs)
        thread.start()
        generated_text = ""
        start = perf_counter()
        times = []
        with torch.inference_mode():
            for new_text in streamer:
                times.append(perf_counter() - start)
                print(new_text, end="", flush=True)
                generated_text += new_text
                start = perf_counter()
        print(f"\n{i}. {generated_text=} TTFT: {times[0]:.2f}s total time: {sum(times):.2f}s")


def cover_chat(chat: list):
    # print(chat)
    audio_cn = 0
    image_cn = 0
    for item in chat:
        tmp_text = ""
        tmp_content = ""
        sub_audio_cn = 0
        sub_image_cn = 0
        if "content" in item and isinstance(item["content"], list):
            for c_item in item["content"]:
                assert isinstance(c_item, dict)
                if "text" in c_item:
                    tmp_text = c_item["text"]
                if "image" in c_item:
                    sub_image_cn += 1
                if "audio" in c_item:
                    sub_audio_cn += 1
            image_cn += sub_image_cn
            audio_cn += sub_audio_cn
            for i in range(image_cn - sub_image_cn, image_cn):
                tmp_content += f"<|image_{i+1}|>"
            for i in range(audio_cn - sub_audio_cn, audio_cn):
                tmp_content += f"<|audio_{i+1}|>"
            if tmp_text:
                tmp_content += tmp_text
            item["content"] = tmp_content

    return chat


def tokenize(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    url = "https://www.ilankelman.org/stopsigns/australia.jpg"
    image = Image.open(requests.get(url, stream=True).raw)
    print(image)
    AUDIO_FILE_1 = os.path.join(model_path, "examples/what_is_the_traffic_sign_in_the_image.wav")
    AUDIO_FILE_2 = os.path.join(model_path, "examples/what_is_shown_in_this_image.wav")
    audio_1 = sf.read(AUDIO_FILE_1)
    audio_2 = sf.read(AUDIO_FILE_2)
    print(f"{audio_1[1]} {audio_1[0].shape=} {audio_1[0].reshape(-1).shape=}")
    print(f"{audio_2[1]} {audio_2[0].shape=} {audio_2[0].reshape(-1).shape=}")
    chat = [
        {"role": "system", "content": [{"type": "text", "text": "You are a helpful assistant."}]},
        {
            "role": "user",
            "content": [
                {"type": "audio", "audio": audio_1[0].reshape(-1)},
                {"type": "image", "image": image},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "The image depicts a street scene with a prominent red stop sign in the foreground. The background showcases a building with traditional Chinese architecture, characterized by its red roof and ornate decorations. There are also several statues of lions, which are common in Chinese culture, positioned in front of the building. The street is lined with various shops and businesses, and there's a car passing by.",
                }
            ],
        },
        {
            "role": "user",
            "content": [{"type": "audio", "audio": audio_2[0].reshape(-1)}],
        },
    ]
    # mm prompt
    audios, images, videos = process_mm_info(chat, use_audio_in_video=False)
    {print(f"audios[{i}]: {item.shape}") for i, item in enumerate(audios)} if audios else print(
        audios
    )
    {print(f"images[{i}]: {item}") for i, item in enumerate(images)} if images else print(images)
    {print(f"videos[{i}]: {item.shape}") for i, item in enumerate(videos)} if videos else print(
        videos
    )

    # text promt
    chat = cover_chat(chat)
    print(chat)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    prompt = processor.tokenizer.apply_chat_template(
        chat, tokenize=False, add_generation_prompt=True
    )
    # need to remove last <|endoftext|> if it is there, which is used for training, not inference. For training, make sure to add <|endoftext|> in the end.
    if prompt.endswith("<|endoftext|>"):
        prompt = prompt.rstrip("<|endoftext|>")
    print(f">>> Prompt\n{prompt}")

    # tokenize (tokens -> token_ids)
    new_audios = []
    for audio in audios:
        new_audios.append((audio, 16000))
    inputs = processor(text=prompt, images=images, audios=new_audios, return_tensors="pt").to(
        "cuda" if gpu_prop else "cpu", dtype=torch.bfloat16
    )
    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]

    # text token
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")


def tokenize_debug(gpu_prop):
    model_name = os.getenv("LLM_MODEL", "microsoft/Phi-4-multimodal-instruct")
    model_path = os.path.join(HF_MODEL_DIR, model_name)
    url = "https://www.ilankelman.org/stopsigns/australia.jpg"
    image = Image.open(requests.get(url, stream=True).raw)
    print(image)
    chat = [
        {"role": "system", "content": "请用中文交流"},
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "你好"},
                {"type": "image", "image": image},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "Hello! Is there something specific you would like to know or discuss?",
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "你叫什么名字"},
                {"type": "image", "image": image},
            ],
        },
    ]
    # mm prompt
    audios, images, videos = process_mm_info(chat, use_audio_in_video=False)
    {print(f"audios[{i}]: {item.shape}") for i, item in enumerate(audios)} if audios else print(
        audios
    )
    {print(f"images[{i}]: {item}") for i, item in enumerate(images)} if images else print(images)
    {print(f"videos[{i}]: {item.shape}") for i, item in enumerate(videos)} if videos else print(
        videos
    )

    # text promt
    chat = cover_chat(chat)
    # print(chat)
    processor = AutoProcessor.from_pretrained(model_path, trust_remote_code=True, use_fast=True)
    prompt = processor.tokenizer.apply_chat_template(
        chat, tokenize=False, add_generation_prompt=True
    )
    # need to remove last <|endoftext|> if it is there, which is used for training, not inference. For training, make sure to add <|endoftext|> in the end.
    if prompt.endswith("<|endoftext|>"):
        prompt = prompt.rstrip("<|endoftext|>")
    print(f"{prompt=}")

    # tokenize (tokens -> token_ids)
    if audios:
        new_audios = []
        for audio in audios:
            new_audios.append((audio, 16000))
        audios = new_audios if new_audios else None
    inputs = processor(text=prompt, images=images, audios=audios, return_tensors="pt").to(
        "cuda" if gpu_prop else "cpu", dtype=torch.bfloat16
    )
    for key, value in inputs.items():
        if value is not None:
            print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]

    # text token
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")


"""
language:
- Text: Arabic, Chinese, Czech, Danish, Dutch, English, Finnish, French, German, Hebrew, Hungarian, Italian, Japanese, Korean, Norwegian, Polish, Portuguese, Russian, Spanish, Swedish, Thai, Turkish, Ukrainian
- Vision: English 
- Audio: English, Chinese, German, French, Italian, Japanese, Spanish, Portuguese

# multimmodal input formats:
- https://huggingface.co/microsoft/Phi-4-multimodal-instruct#input-formats

modal run src/download_models.py --repo-ids "microsoft/Phi-4-mini-instruct,microsoft/Phi-4-multimodal-instruct"

IMAGE_GPU=T4 modal run src/llm/transformers/phi4_multimodal.py --task tokenize_debug
IMAGE_GPU=T4 modal run src/llm/transformers/phi4_multimodal.py --task tokenize
IMAGE_GPU=T4 modal run src/llm/transformers/phi4_multimodal.py --task dump_model

IMAGE_GPU=T4 modal run src/llm/transformers/phi4_multimodal.py --task text2text
IMAGE_GPU=T4 modal run src/llm/transformers/phi4_multimodal.py --task text2text_stream

# NOTE: more vision(image) + speech encode need more memory
IMAGE_GPU=T4 modal run src/llm/transformers/phi4_multimodal.py --task asr
IMAGE_GPU=T4 modal run src/llm/transformers/phi4_multimodal.py --task multi_asr
IMAGE_GPU=T4 modal run src/llm/transformers/phi4_multimodal.py --task single_vision2text 
IMAGE_GPU=L4 modal run src/llm/transformers/phi4_multimodal.py --task multi_vision2text 
IMAGE_GPU=T4 modal run src/llm/transformers/phi4_multimodal.py --task single_turn_vision2text
IMAGE_GPU=L4 modal run src/llm/transformers/phi4_multimodal.py --task multi_turn_vision2text 
IMAGE_GPU=L4 modal run src/llm/transformers/phi4_multimodal.py --task speech_vision2text
IMAGE_GPU=L4 modal run src/llm/transformers/phi4_multimodal.py --task multi_turn_vision_speech2text
IMAGE_GPU=L4 modal run src/llm/transformers/phi4_multimodal.py --task multi_turn_vision_speech2text_stream
"""


@app.local_entrypoint()
def main(task: str = "dump_model"):
    tasks = {
        "dump_model": dump_model,
        "tokenize_debug": tokenize_debug,
        "tokenize": tokenize,
        "text2text": text2text,
        "text2text_stream": text2text_stream,
        "single_vision2text": single_vision2text,
        "multi_vision2text": multi_vision2text,
        "single_turn_vision2text": single_turn_vision2text,
        "multi_turn_vision2text": multi_turn_vision2text,
        "speech_vision2text": speech_vision2text,
        "asr": asr,
        "multi_asr": multi_asr,
        "multi_turn_vision_speech2text": multi_turn_vision_speech2text,
        "multi_turn_vision_speech2text_stream": multi_turn_vision_speech2text_stream,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    run.remote(tasks[task])



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""isort:skip_file"""

import argparse
import importlib
import os

from fairseq.dataclass import FairseqDataclass
from fairseq.dataclass.utils import merge_with_parent, populate_dataclass
from hydra.core.config_store import ConfigStore

from .composite_encoder import CompositeEncoder
from .distributed_fairseq_model import DistributedFairseqModel
from .fairseq_decoder import FairseqDecoder
from .fairseq_encoder import FairseqEncoder
from .fairseq_incremental_decoder import FairseqIncrementalDecoder
from .fairseq_model import (
    BaseFairseqModel,
    FairseqEncoderDecoderModel,
    FairseqEncoderModel,
    FairseqLanguageModel,
    FairseqModel,
    FairseqMultiModel,
)


MODEL_REGISTRY = {}
MODEL_DATACLASS_REGISTRY = {}
ARCH_MODEL_REGISTRY = {}
ARCH_MODEL_NAME_REGISTRY = {}
ARCH_MODEL_INV_REGISTRY = {}
ARCH_CONFIG_REGISTRY = {}


__all__ = [
    "BaseFairseqModel",
    "CompositeEncoder",
    "DistributedFairseqModel",
    "FairseqDecoder",
    "FairseqEncoder",
    "FairseqEncoderDecoderModel",
    "FairseqEncoderModel",
    "FairseqIncrementalDecoder",
    "FairseqLanguageModel",
    "FairseqModel",
    "FairseqMultiModel",
]


def build_model(cfg: FairseqDataclass, task):

    model = None
    model_type = getattr(cfg, "_name", None) or getattr(cfg, "arch", None)
    if model_type == "wav2vec_mtls1":
        model_type = "unispeech"
    if 'wav2vec2' in model_type:
        model_type = 'wav2vec2'

    if not model_type and len(cfg) == 1:
        # this is hit if config object is nested in directory that is named after model type

        model_type = next(iter(cfg))
        if model_type in MODEL_DATACLASS_REGISTRY:
            cfg = cfg[model_type]
        else:
            raise Exception(
                "Could not infer model type from directory. Please add _name field to indicate model type. "
                "Available models: "
                + str(MODEL_DATACLASS_REGISTRY.keys())
                + " Requested model type: "
                + model_type
            )

    if model_type in ARCH_MODEL_REGISTRY:
        # case 1: legacy models
        model = ARCH_MODEL_REGISTRY[model_type]
    elif model_type in MODEL_DATACLASS_REGISTRY:
        # case 2: config-driven models
        model = MODEL_REGISTRY[model_type]

    if model_type in MODEL_DATACLASS_REGISTRY:
        # set defaults from dataclass. note that arch name and model name can be the same
        dc = MODEL_DATACLASS_REGISTRY[model_type]
        if isinstance(cfg, argparse.Namespace):
            cfg = populate_dataclass(dc(), cfg)
        else:
            cfg = merge_with_parent(dc(), cfg)

    assert model is not None, (
        f"Could not infer model type from {cfg}. "
        f"Available models: "
        + str(MODEL_DATACLASS_REGISTRY.keys())
        + " Requested model type: "
        + model_type
    )

    return model.build_model(cfg, task)


def register_model(name, dataclass=None):
    """
    New model types can be added to fairseq with the :func:`register_model`
    function decorator.

    For example::

        @register_model('lstm')
        class LSTM(FairseqEncoderDecoderModel):
            (...)

    .. note:: All models must implement the :class:`BaseFairseqModel` interface.
        Typically you will extend :class:`FairseqEncoderDecoderModel` for
        sequence-to-sequence tasks or :class:`FairseqLanguageModel` for
        language modeling tasks.

    Args:
        name (str): the name of the model
    """

    def register_model_cls(cls):
        if name in MODEL_REGISTRY:
            raise ValueError("Cannot register duplicate model ({})".format(name))
        if not issubclass(cls, BaseFairseqModel):
            raise ValueError(
                "Model ({}: {}) must extend BaseFairseqModel".format(name, cls.__name__)
            )
        MODEL_REGISTRY[name] = cls
        if dataclass is not None and not issubclass(dataclass, FairseqDataclass):
            raise ValueError(
                "Dataclass {} must extend FairseqDataclass".format(dataclass)
            )

        cls.__dataclass = dataclass
        if dataclass is not None:
            MODEL_DATACLASS_REGISTRY[name] = dataclass

            cs = ConfigStore.instance()
            node = dataclass()
            node._name = name
            cs.store(name=name, group="model", node=node, provider="fairseq")

            @register_model_architecture(name, name)
            def noop(_):
                pass

        return cls

    return register_model_cls


def register_model_architecture(model_name, arch_name):
    """
    New model architectures can be added to fairseq with the
    :func:`register_model_architecture` function decorator. After registration,
    model architectures can be selected with the ``--arch`` command-line
    argument.

    For example::

        @register_model_architecture('lstm', 'lstm_luong_wmt_en_de')
        def lstm_luong_wmt_en_de(cfg):
            args.encoder_embed_dim = getattr(cfg.model, 'encoder_embed_dim', 1000)
            (...)

    The decorated function should take a single argument *cfg*, which is a
    :class:`omegaconf.DictConfig`. The decorated function should modify these
    arguments in-place to match the desired architecture.

    Args:
        model_name (str): the name of the Model (Model must already be
            registered)
        arch_name (str): the name of the model architecture (``--arch``)
    """

    def register_model_arch_fn(fn):
        if model_name not in MODEL_REGISTRY:
            raise ValueError(
                "Cannot register model architecture for unknown model type ({})".format(
                    model_name
                )
            )
        if arch_name in ARCH_MODEL_REGISTRY:
            raise ValueError(
                "Cannot register duplicate model architecture ({})".format(arch_name)
            )
        if not callable(fn):
            raise ValueError(
                "Model architecture must be callable ({})".format(arch_name)
            )
        ARCH_MODEL_REGISTRY[arch_name] = MODEL_REGISTRY[model_name]
        ARCH_MODEL_NAME_REGISTRY[arch_name] = model_name
        ARCH_MODEL_INV_REGISTRY.setdefault(model_name, []).append(arch_name)
        ARCH_CONFIG_REGISTRY[arch_name] = fn
        return fn

    return register_model_arch_fn


def import_models(models_dir, namespace):
    for file in os.listdir(models_dir):
        path = os.path.join(models_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            model_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + model_name)

            # extra `model_parser` for sphinx
            if model_name in MODEL_REGISTRY:
                parser = argparse.ArgumentParser(add_help=False)
                group_archs = parser.add_argument_group("Named architectures")
                group_archs.add_argument(
                    "--arch", choices=ARCH_MODEL_INV_REGISTRY[model_name]
                )
                group_args = parser.add_argument_group(
                    "Additional command-line arguments"
                )
                MODEL_REGISTRY[model_name].add_args(group_args)
                globals()[model_name + "_parser"] = parser


# automatically import any Python files in the models/ directory
models_dir = os.path.dirname(__file__)
import_models(models_dir, "fairseq.models")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


class DynamicLossScaler(object):
    def __init__(
        self,
        init_scale=2.0 ** 15,
        scale_factor=2.0,
        scale_window=2000,
        tolerance=0.0,
        threshold=None,
        min_loss_scale=1e-4,
    ):
        self.loss_scale = init_scale
        self.scale_factor = scale_factor
        self.scale_window = scale_window
        self.tolerance = tolerance
        self.threshold = threshold
        self._iter = 0
        self._last_overflow_iter = -1
        self._last_rescale_iter = -1
        self._overflows_since_rescale = 0
        self.min_loss_scale = min_loss_scale

    def scale(self, outputs):
        return self.loss_scale * outputs

    def update(self):
        if (self._iter - self._last_overflow_iter) % self.scale_window == 0:
            self.loss_scale *= self.scale_factor
            self._last_rescale_iter = self._iter
        self._iter += 1

    def _decrease_loss_scale(self):
        self.loss_scale /= self.scale_factor
        if self.threshold is not None:
            self.loss_scale = max(self.loss_scale, self.threshold)

    def check_overflow(self, grad_norm):
        # detect inf and nan
        if grad_norm == float("inf") or grad_norm != grad_norm:
            # overflow has occured
            prev_scale = self.loss_scale
            iter_since_rescale = self._iter - self._last_rescale_iter

            self._last_overflow_iter = self._iter
            self._overflows_since_rescale += 1
            pct_overflow = self._overflows_since_rescale / float(iter_since_rescale)
            if pct_overflow >= self.tolerance:
                self._decrease_loss_scale()
                self._last_rescale_iter = self._iter
                self._overflows_since_rescale = 0

            if self.loss_scale <= self.min_loss_scale:
                # Use FloatingPointError as an uncommon error that parent
                # functions can safely catch to stop training.
                self.loss_scale = prev_scale
                raise FloatingPointError(
                    (
                        "Minimum loss scale reached ({}). Your loss is probably exploding. "
                        "Try lowering the learning rate, using gradient clipping or "
                        "increasing the batch size."
                    ).format(self.min_loss_scale)
                )

            self._iter += 1
            raise OverflowError("setting loss scale to: " + str(self.loss_scale))



# Copyright (c) 2023-2024 DeepSeek.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import json
from typing import Dict, List

import PIL.Image
import torch
import base64
import io
from transformers import AutoModelForCausalLM

from janus.models import MultiModalityCausalLM, VLChatProcessor


def load_pretrained_model(model_path: str):
    vl_chat_processor: VLChatProcessor = VLChatProcessor.from_pretrained(model_path)
    tokenizer = vl_chat_processor.tokenizer

    vl_gpt: MultiModalityCausalLM = AutoModelForCausalLM.from_pretrained(
        model_path, trust_remote_code=True
    )
    vl_gpt = vl_gpt.to(torch.bfloat16).cuda().eval()

    return tokenizer, vl_chat_processor, vl_gpt


def load_pil_images(conversations: List[Dict[str, str]]) -> List[PIL.Image.Image]:
    """

    Support file path or base64 images.

    Args:
        conversations (List[Dict[str, str]]): the conversations with a list of messages. An example is :
            [
                {
                    "role": "User",
                    "content": "<image_placeholder>\nExtract all information from this image and convert them into markdown format.",
                    "images": ["./examples/table_datasets.png"]
                },
                {"role": "Assistant", "content": ""},
            ]

    Returns:
        pil_images (List[PIL.Image.Image]): the list of PIL images.

    """

    pil_images = []

    for message in conversations:
        if "images" not in message:
            continue

        for image_data in message["images"]:
            if image_data.startswith("data:image"):
                # Image data is in base64 format
                _, image_data = image_data.split(",", 1)
                image_bytes = base64.b64decode(image_data)
                pil_img = PIL.Image.open(io.BytesIO(image_bytes))
            else:
                # Image data is a file path
                pil_img = PIL.Image.open(image_data)
            pil_img = pil_img.convert("RGB")
            pil_images.append(pil_img)

    return pil_images


def load_json(filepath):
    with open(filepath, "r") as f:
        data = json.load(f)
        return data



import os
import logging

import unittest
from apipeline.pipeline.pipeline import Pipeline, FrameDirection
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.pipeline.runner import PipelineRunner
from apipeline.frames.data_frames import DataFrame, TextFrame, Frame
from apipeline.frames.control_frames import EndFrame
from apipeline.processors.frame_processor import FrameProcessor

from src.modules.speech.asr import ASREnvInit
from src.modules.speech.vad_analyzer.silero import SileroVADAnalyzer
from src.processors.speech.asr.asr_processor import ASRProcessor
from src.common.types import DailyParams
from src.common.logger import Logger
from src.transports.daily import DailyTransport
from src.types.frames.data_frames import TranscriptionFrame
from src.common.session import Session
from src.common.types import SessionCtx

from dotenv import load_dotenv

load_dotenv(override=True)

"""
ASR_TAG="whisper_asr" \
    python -m unittest test.integration.processors.test_asr_processor.TestASRProcessor
ASR_TAG="whisper_timestamped_asr" \
    python -m unittest test.integration.processors.test_asr_processor.TestASRProcessor
ASR_TAG="whisper_faster_asr" \
    python -m unittest test.integration.processors.test_asr_processor.TestASRProcessor
ASR_TAG="whisper_transformers_asr" \
    ASR_MODEL_NAME_OR_PATH=./models/openai/whisper-base \
    python -m unittest test.integration.processors.test_asr_processor.TestASRProcessor
ASR_TAG="whisper_mlx_asr" \
    python -m unittest test.integration.processors.test_asr_processor.TestASRProcessor
ASR_TAG="sense_voice_asr" \
    ASR_LANG=zn \
    ASR_MODEL_NAME_OR_PATH=./models/FunAudioLLM/SenseVoiceSmall \
    python -m unittest test.integration.processors.test_asr_processor.TestASRProcessor
ASR_TAG="whisper_groq_asr" \
    ASR_LANG=zh \
    ASR_MODEL_NAME_OR_PATH=whisper-large-v3 \
    python -m unittest test.integration.processors.test_asr_processor.TestASRProcessor

ASR_TAG="whisper_groq_asr" \
    ASR_LANG=zh \
    ASR_MODEL_NAME_OR_PATH=whisper-large-v3 \
    DAILY_VAD_ENABLED=1 \
    python -m unittest test.integration.processors.test_asr_processor.TestASRProcessor
"""


class TranscriptionLogger(FrameProcessor):
    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, TranscriptionFrame):
            print(f"get Transcription Frame: {frame}, text len:{len(frame.text)}")


class TestASRProcessor(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False)
        cls.room_url = os.getenv("DAILY_ROOM_URL", "https://weedge.daily.co/chat-room")
        cls.room_token = os.getenv("DAILY_ROOM_TOKEN", None)
        cls.vad_enabled = bool(os.getenv("DAILY_VAD_ENABLED", ""))

    @classmethod
    def tearDownClass(cls):
        pass

    async def asyncSetUp(self):
        daily_pramams: DailyParams = DailyParams(
            audio_in_enabled=True,
            transcription_enabled=False,
        )
        if self.vad_enabled is True:
            daily_pramams.vad_enabled = True
            daily_pramams.vad_analyzer = SileroVADAnalyzer()
            daily_pramams.vad_audio_passthrough = True
        transport = DailyTransport(
            self.room_url,
            self.room_token,
            "Transcription bot",
            daily_pramams,
        )

        asr = ASREnvInit.initASREngine()
        session = Session(**SessionCtx("test_client_id", 16000, 2).__dict__)
        asr_processor = ASRProcessor(asr=asr, session=session)

        tl_porcessor = TranscriptionLogger()

        pipeline = Pipeline([transport.input_processor(), asr_processor, tl_porcessor])

        self.task = PipelineTask(
            pipeline,
            PipelineParams(
                allow_interruptions=True,
                enable_metrics=True,
            ),
        )

    async def asyncTearDown(self):
        pass

    async def test_asr(self):
        runner = PipelineRunner()
        await runner.run(self.task)



"""
Usage:
python3 -m llava.model.consolidate --src ~/model_weights/llava-7b --dst ~/model_weights/llava-7b_consolidate
"""
import argparse

import torch
from transformers import AutoTokenizer, AutoModelForCausalLM
from llava.model import *
from llava.model.utils import auto_upgrade


def consolidate_ckpt(src_path, dst_path):
    print("Loading model")
    auto_upgrade(src_path)
    src_model = AutoModelForCausalLM.from_pretrained(src_path, torch_dtype=torch.float16, low_cpu_mem_usage=True)
    src_tokenizer = AutoTokenizer.from_pretrained(src_path, use_fast=False)
    src_model.save_pretrained(dst_path)
    src_tokenizer.save_pretrained(dst_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--src", type=str, required=True)
    parser.add_argument("--dst", type=str, required=True)

    args = parser.parse_args()

    consolidate_ckpt(args.src, args.dst)



import os
from typing import List

import tqdm
import torch
from loguru import logger
from huggingface_hub import cached_assets_path
from transformers import AutoModelForCausalLM

from kimia_infer.models.detokenizer import get_audio_detokenizer
from .prompt_manager import KimiAPromptManager
from kimia_infer.utils.sampler import KimiASampler
from huggingface_hub import snapshot_download

class KimiAudio(object):
    def __init__(self, model_path: str, load_detokenizer: bool = True):
        logger.info(f"Loading kimi-audio main model")

        if os.path.exists(model_path):
            # local path
            cache_path = model_path
        else:
            # cache everything if model_path is a model-id
            cache_path = snapshot_download(model_path)
    
        logger.info(f"Looking for resources in {cache_path}")
        logger.info(f"Loading whisper model")
        self.alm = AutoModelForCausalLM.from_pretrained(
            cache_path, torch_dtype=torch.bfloat16, trust_remote_code=True
        )
        self.alm = self.alm.to(torch.cuda.current_device())

        model_config = self.alm.config
        self.kimia_token_offset = model_config.kimia_token_offset

        self.prompt_manager = KimiAPromptManager(
            model_path=cache_path, kimia_token_offset=self.kimia_token_offset
        )

        if load_detokenizer:
            logger.info(f"Loading detokenizer")
            # need to compile extension moudules for the first time, it may take several minutes.
            self.detokenizer = get_audio_detokenizer(cache_path)
        else:
            # in this case, you're not allowed to generate audio(wav)
            self.detokenizer = None

        self.extra_tokens = self.prompt_manager.extra_tokens
        self.kimia_text_audiodelaytokens = 6
        self.eod_ids = [self.extra_tokens.msg_end, self.extra_tokens.media_end]

    @torch.inference_mode()
    def _generate_loop(
        self,
        audio_input_ids: torch.Tensor,  # input audio tokens
        text_input_ids: torch.Tensor = None,  # input text tokens if use multi-input
        max_new_tokens: int = 50,
        audio_top_k: int = 5,
        audio_temperature: float = 0.0,
        audio_repetition_penalty: float = 1.0,
        audio_repetition_window_size: int = 64,
        text_top_k: int = 5,
        text_temperature: float = 0.0,
        text_repetition_penalty: float = 1.0,
        text_repetition_window_size: int = 16,
        is_continuous_mask: torch.Tensor = None,
        continous_feature: List[torch.Tensor] = [],
        output_type: str = "text",
    ):

        sampler = KimiASampler(
            audio_top_k=audio_top_k,
            audio_temperature=audio_temperature,
            audio_repetition_penalty=audio_repetition_penalty,
            audio_repetition_window_size=audio_repetition_window_size,
            text_top_k=text_top_k,
            text_temperature=text_temperature,
            text_repetition_penalty=text_repetition_penalty,
            text_repetition_window_size=text_repetition_window_size,
        )

        text_stream_is_finished = False
        previous_audio_tokens = torch.zeros(
            (4096,),
            dtype=torch.int,
            device=torch.cuda.current_device(),
        )
        text_previous_tokens = torch.zeros(
            (4096,),
            dtype=torch.int,
            device=torch.cuda.current_device(),
        )

        decoder_input_audio_ids = audio_input_ids.clone()
        decoder_input_text_ids = text_input_ids.clone()
        decoder_position_ids = (
            torch.arange(
                0, decoder_input_audio_ids.shape[1], device=torch.cuda.current_device()
            )
            .unsqueeze(0)
            .long()
        )
        decoder_input_whisper_feature = continous_feature or None
        decoder_is_continuous_mask = is_continuous_mask
        past_key_values = None

        last_position_id = decoder_input_audio_ids.shape[1] - 1

        valid_text_length = 0
        valid_audio_length = 0

        for i in tqdm.tqdm(range(max_new_tokens), desc="Generating tokens"):
            # https://huggingface.co/moonshotai/Kimi-Audio-7B-Instruct/blob/main/modeling_moonshot_kimia.py#L850
            audio_logits, text_logits, past_key_values = self.alm.forward(
                input_ids=decoder_input_audio_ids,
                text_input_ids=decoder_input_text_ids,
                whisper_input_feature=decoder_input_whisper_feature,
                is_continuous_mask=decoder_is_continuous_mask,
                position_ids=decoder_position_ids,
                past_key_values=past_key_values,
                return_dict=False,
            )

            # Sample text token using the sampler
            next_token_text = sampler.sample_text_logits(
                text_logits, recent_tokens=text_previous_tokens[:i] if i > 0 else None
            )

            # Sample audio token using the sampler
            next_audio_token = sampler.sample_audio_logits(
                audio_logits, recent_tokens=previous_audio_tokens[:i] if i > 0 else None
            )

            if text_stream_is_finished:
                next_token_text.fill_(self.extra_tokens.kimia_text_blank)
            elif next_token_text.item() == self.extra_tokens.kimia_text_eos:
                text_stream_is_finished = True
            else:
                valid_text_length += 1

            text_previous_tokens[i : i + 1] = next_token_text

            if i < self.kimia_text_audiodelaytokens:
                next_audio_token.fill_(self.extra_tokens.kimia_text_blank)
            else:
                if output_type == "text":
                    next_audio_token.fill_(self.extra_tokens.kimia_text_blank)
                else:
                    valid_audio_length += 1

            previous_audio_tokens[i : i + 1] = next_audio_token

            audio_stream_is_finished = next_audio_token.item() in self.eod_ids

            if (
                output_type == "text"
                and text_stream_is_finished
                or output_type == "both"
                and audio_stream_is_finished
            ):
                return_text_tokens = (
                    text_previous_tokens[:valid_text_length]
                    .detach()
                    .cpu()
                    .numpy()
                    .tolist()
                )
                return_audio_tokens = (
                    previous_audio_tokens[
                        self.kimia_text_audiodelaytokens : valid_audio_length
                        + self.kimia_text_audiodelaytokens
                    ]
                    .detach()
                    .cpu()
                    .numpy()
                    .tolist()
                )
                return return_audio_tokens, return_text_tokens
            else:
                decoder_input_audio_ids = next_audio_token.unsqueeze(1)
                decoder_input_text_ids = next_token_text.unsqueeze(1)

                decoder_position_ids = (
                    torch.zeros(1, 1, device=torch.cuda.current_device())
                    .fill_(last_position_id + 1)
                    .long()
                    .view(1, 1)
                )
                last_position_id += 1

                decoder_input_whisper_feature = None
                decoder_is_continuous_mask = None

        return_text_tokens = (
            text_previous_tokens[:valid_text_length].detach().cpu().numpy().tolist()
        )
        return_audio_tokens = (
            previous_audio_tokens[
                self.kimia_text_audiodelaytokens : valid_audio_length
                + self.kimia_text_audiodelaytokens
            ]
            .detach()
            .cpu()
            .numpy()
            .tolist()
        )
        return return_audio_tokens, return_text_tokens

    @torch.inference_mode()
    def generate(
        self,
        chats: list[dict],
        output_type="text",
        audio_temperature=0.0,
        audio_top_k=5,
        text_temperature=0.0,
        text_top_k=5,
        audio_repetition_penalty=1.0,
        audio_repetition_window_size=64,
        text_repetition_penalty=1.0,
        text_repetition_window_size=16,
        max_new_tokens=-1,
    ):
        ## TODO: 需要一个check函数，检查输入的history格式是否合法
        ## 比如，对于ASR任务，一定是: text-instruction/audio-instruction + audio-content, 我理解content和instruction是不能换位置的
        ## assistant前必须有user等等，我觉得最好做一下check

        assert output_type in ["text", "both"]

        history = self.prompt_manager.get_prompt(chats, output_type=output_type)

        audio_input_ids, text_input_ids, is_continuous_mask = history.to_tensor()
        audio_features = history.continuous_feature

        generated_wav_tokens = []
        generated_text_tokens = []

        if output_type == "both":
            max_new_tokens = int(12.5 * 120) - audio_input_ids.shape[1]
        else:
            if max_new_tokens == -1:
                max_new_tokens = 7500 - audio_input_ids.shape[1]

        audio_input_ids = audio_input_ids.to(torch.cuda.current_device())
        text_input_ids = text_input_ids.to(torch.cuda.current_device())
        is_continuous_mask = is_continuous_mask.to(torch.cuda.current_device())
        audio_features = [f.to(torch.cuda.current_device()) for f in audio_features]

        generated_wav_tokens, generated_text_tokens = self._generate_loop(
            audio_input_ids=audio_input_ids,
            text_input_ids=text_input_ids,
            max_new_tokens=max_new_tokens,
            audio_temperature=audio_temperature,
            audio_top_k=audio_top_k,
            audio_repetition_penalty=audio_repetition_penalty,
            audio_repetition_window_size=audio_repetition_window_size,
            text_top_k=text_top_k,
            text_temperature=text_temperature,
            text_repetition_penalty=text_repetition_penalty,
            text_repetition_window_size=text_repetition_window_size,
            is_continuous_mask=is_continuous_mask,
            continous_feature=audio_features,
            output_type=output_type,
        )
        generated_wav_tokens = [
            t for t in generated_wav_tokens if t >= self.kimia_token_offset
        ]  #  filter out the illegal tokens

        generated_wav_tokens = torch.tensor(generated_wav_tokens).unsqueeze(0)
        generated_wav_tokens = generated_wav_tokens - self.kimia_token_offset

        generated_text_tokens = [
            t for t in generated_text_tokens if t < self.kimia_token_offset
        ]
        generated_text = self.detokenize_text(generated_text_tokens)
        if self.detokenizer is not None and output_type == "both":
            generated_wav = self.detokenize_audio(generated_wav_tokens)
        else:
            generated_wav = None

        return generated_wav, generated_text

    def detokenize_audio(self, audio_tokens):
        if self.detokenizer is None:
            raise ValueError("Detokenizer is not initialized")
        self.detokenizer.clear_states()
        chunk_size = 30  # hard-coded right now
        first_chunk_size = 30
        cache_speech_collection = []
        audio_tokens = audio_tokens.to(torch.cuda.current_device())
        audio_tokens = audio_tokens.long()
        num_audio_tokens = audio_tokens.size(1)
        first_chunk_semantic_tokens = audio_tokens[:, :first_chunk_size]
        gen_speech = self.detokenizer.detokenize_streaming(
            first_chunk_semantic_tokens,
            is_final=(num_audio_tokens <= first_chunk_size),
            upsample_factor=4,
        )
        cache_speech_collection.append(gen_speech)

        if num_audio_tokens > first_chunk_size:
            res_semantic_tokens = audio_tokens[:, first_chunk_size:]
            for i in range(0, res_semantic_tokens.size(1), chunk_size):
                chunk_semantic_tokens = res_semantic_tokens[:, i : i + chunk_size]
                gen_speech = self.detokenizer.detokenize_streaming(
                    chunk_semantic_tokens,
                    upsample_factor=4,
                    is_final=(i + chunk_size >= res_semantic_tokens.size(1)),
                )
                cache_speech_collection.append(gen_speech)

        gen_speech = torch.cat(cache_speech_collection, dim=-1)
        return gen_speech

    def detokenize_text(self, text_tokens):
        valid_text_ids = []
        for x in text_tokens:
            if x == self.extra_tokens.kimia_text_eos:
                break
            valid_text_ids.append(x)
        return self.prompt_manager.text_tokenizer.decode(valid_text_ids)



from mcp.shared._httpx_utils import create_mcp_http_client
from mcp.shared.context import RequestContext
import mcp.types as types

from .tool_register import functions


@functions.register("fetch_website")
async def fetch_website(
    ctx: RequestContext,
    url: str,
) -> list[types.TextContent | types.ImageContent | types.EmbeddedResource]:
    headers = {"User-Agent": "MCP Test Server (github.com/modelcontextprotocol/python-sdk)"}
    async with create_mcp_http_client(headers=headers) as client:
        response = await client.get(url)
        response.raise_for_status()
        return [types.TextContent(type="text", text=response.text)]



import logging

import click

from . import mcp
from .search import *


@click.command()
@click.option(
    "--host",
    default="127.0.0.1",
    help="Host to listen on for SSE or streamable-http",
)
@click.option(
    "--port",
    default=8000,
    help="Port to listen on for SSE or streamable-http",
)
@click.option(
    "--transport",
    type=click.Choice(
        [
            "stdio",
            "sse",
            "streamable-http",
        ]
    ),
    default="stdio",
    help="Transport type",
)
@click.option(
    "--log-level",
    default="INFO",
    help="Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
)
@click.option(
    "--json-response",
    is_flag=True,
    default=False,
    help="Enable JSON responses instead of SSE streams",
)
@click.option(
    "--stateless-http",
    is_flag=True,
    default=False,
    help="Enable stateless HTTP transport",
)
def main(
    host: str,
    port: int,
    transport: str,
    log_level: str,
    json_response: bool,
    stateless_http: bool,
) -> int:
    # Configure logging
    logging.basicConfig(
        level=getattr(logging, log_level.upper()),
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )
    mcp.settings.json_response = json_response
    mcp.settings.stateless_http = stateless_http
    logging.info(f"Starting MCP Server, {mcp.settings=}")

    if transport == "streamable-http":
        mcp.run(
            transport=transport,
            host=host,
            port=port,
            path="/mcp",
            log_level=log_level.lower(),
        )
    elif transport == "sse":
        mcp.run(
            transport=transport,
            host=host,
            port=port,
            log_level=log_level.lower(),
        )
    else:
        mcp.run(transport=transport)



import os
import sys

def align_trans_scp_file(trans, scp):
    trans_dict = {}
    with open(trans, 'r') as f:
        for line in f:
            sec = line.strip().split("\t")
            trans_dict[sec[0]] = sec[1]
    scp_dict = {}
    with open(scp, 'r') as f:
        for line in f:
            sec = line.strip().split(" ")
            scp_dict[sec[0]] = sec[1]
    with open("text", "w") as f:
        for k, v in scp_dict.items():
            f.write("%s\t%s\n"%(k,trans_dict[k]))

if __name__ == '__main__':
    trans = sys.argv[1]
    scp = sys.argv[2]
    align_trans_scp_file(trans, scp)


# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""EnCodec model implementation."""

import math
from pathlib import Path
import typing as tp

import numpy as np
import torch
from torch import nn

from . import quantization as qt
from . import modules as m
from .utils import _check_checksum, _linear_overlap_add, _get_checkpoint_url


ROOT_URL = 'https://dl.fbaipublicfiles.com/encodec/v0/'

EncodedFrame = tp.Tuple[torch.Tensor, tp.Optional[torch.Tensor]]


class LMModel(nn.Module):
    """Language Model to estimate probabilities of each codebook entry.
    We predict all codebooks in parallel for a given time step.

    Args:
        n_q (int): number of codebooks.
        card (int): codebook cardinality.
        dim (int): transformer dimension.
        **kwargs: passed to `encoder.modules.transformer.StreamingTransformerEncoder`.
    """
    def __init__(self, n_q: int = 32, card: int = 1024, dim: int = 200, **kwargs):
        super().__init__()
        self.card = card
        self.n_q = n_q
        self.dim = dim
        self.transformer = m.StreamingTransformerEncoder(dim=dim, **kwargs)
        self.emb = nn.ModuleList([nn.Embedding(card + 1, dim) for _ in range(n_q)])
        self.linears = nn.ModuleList([nn.Linear(dim, card) for _ in range(n_q)])

    def forward(self, indices: torch.Tensor,
                states: tp.Optional[tp.List[torch.Tensor]] = None, offset: int = 0):
        """
        Args:
            indices (torch.Tensor): indices from the previous time step. Indices
                should be 1 + actual index in the codebook. The value 0 is reserved for
                when the index is missing (i.e. first time step). Shape should be
                `[B, n_q, T]`.
            states: state for the streaming decoding.
            offset: offset of the current time step.

        Returns a 3-tuple `(probabilities, new_states, new_offset)` with probabilities
        with a shape `[B, card, n_q, T]`.

        """
        B, K, T = indices.shape
        input_ = sum([self.emb[k](indices[:, k]) for k in range(K)])
        out, states, offset = self.transformer(input_, states, offset)
        logits = torch.stack([self.linears[k](out) for k in range(K)], dim=1).permute(0, 3, 1, 2)
        return torch.softmax(logits, dim=1), states, offset


class EncodecModel(nn.Module):
    """EnCodec model operating on the raw waveform.
    Args:
        target_bandwidths (list of float): Target bandwidths.
        encoder (nn.Module): Encoder network.
        decoder (nn.Module): Decoder network.
        sample_rate (int): Audio sample rate.
        channels (int): Number of audio channels.
        normalize (bool): Whether to apply audio normalization.
        segment (float or None): segment duration in sec. when doing overlap-add.
        overlap (float): overlap between segment, given as a fraction of the segment duration.
        name (str): name of the model, used as metadata when compressing audio.
    """
    def __init__(self,
                 encoder: m.SEANetEncoder,
                 decoder: m.SEANetDecoder,
                 quantizer: qt.ResidualVectorQuantizer,
                 target_bandwidths: tp.List[float],
                 sample_rate: int,
                 channels: int,
                 normalize: bool = False,
                 segment: tp.Optional[float] = None,
                 overlap: float = 0.01,
                 name: str = 'unset'):
        super().__init__()
        self.bandwidth: tp.Optional[float] = None
        self.target_bandwidths = target_bandwidths
        self.encoder = encoder
        self.quantizer = quantizer
        self.decoder = decoder
        self.sample_rate = sample_rate
        self.channels = channels
        self.normalize = normalize
        self.segment = segment
        self.overlap = overlap
        self.frame_rate = math.ceil(self.sample_rate / np.prod(self.encoder.ratios))
        self.name = name
        self.bits_per_codebook = int(math.log2(self.quantizer.bins))
        assert 2 ** self.bits_per_codebook == self.quantizer.bins, \
            "quantizer bins must be a power of 2."

    @property
    def segment_length(self) -> tp.Optional[int]:
        if self.segment is None:
            return None
        return int(self.segment * self.sample_rate)

    @property
    def segment_stride(self) -> tp.Optional[int]:
        segment_length = self.segment_length
        if segment_length is None:
            return None
        return max(1, int((1 - self.overlap) * segment_length))

    def encode(self, x: torch.Tensor) -> tp.List[EncodedFrame]:
        """Given a tensor `x`, returns a list of frames containing
        the discrete encoded codes for `x`, along with rescaling factors
        for each segment, when `self.normalize` is True.

        Each frames is a tuple `(codebook, scale)`, with `codebook` of
        shape `[B, K, T]`, with `K` the number of codebooks.
        """
        assert x.dim() == 3
        _, channels, length = x.shape
        assert channels > 0 and channels <= 2
        segment_length = self.segment_length
        if segment_length is None:
            segment_length = length
            stride = length
        else:
            stride = self.segment_stride  # type: ignore
            assert stride is not None

        encoded_frames: tp.List[EncodedFrame] = []
        for offset in range(0, length, stride):
            frame = x[:, :, offset: offset + segment_length]
            encoded_frames.append(self._encode_frame(frame))
        return encoded_frames

    def _encode_frame(self, x: torch.Tensor) -> EncodedFrame:
        length = x.shape[-1]
        duration = length / self.sample_rate
        assert self.segment is None or duration <= 1e-5 + self.segment

        if self.normalize:
            mono = x.mean(dim=1, keepdim=True)
            volume = mono.pow(2).mean(dim=2, keepdim=True).sqrt()
            scale = 1e-8 + volume
            x = x / scale
            scale = scale.view(-1, 1)
        else:
            scale = None

        emb = self.encoder(x)
        codes = self.quantizer.encode(emb, self.frame_rate, self.bandwidth)
        codes = codes.transpose(0, 1)
        # codes is [B, K, T], with T frames, K nb of codebooks.
        return codes, scale

    def decode(self, encoded_frames: tp.List[EncodedFrame]) -> torch.Tensor:
        """Decode the given frames into a waveform.
        Note that the output might be a bit bigger than the input. In that case,
        any extra steps at the end can be trimmed.
        """
        segment_length = self.segment_length
        if segment_length is None:
            assert len(encoded_frames) == 1
            return self._decode_frame(encoded_frames[0])

        frames = [self._decode_frame(frame) for frame in encoded_frames]
        return _linear_overlap_add(frames, self.segment_stride or 1)

    def _decode_frame(self, encoded_frame: EncodedFrame) -> torch.Tensor:
        codes, scale = encoded_frame
        codes = codes.transpose(0, 1)
        emb = self.quantizer.decode(codes)
        out = self.decoder(emb)
        if scale is not None:
            out = out * scale.view(-1, 1, 1)
        return out

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        frames = self.encode(x)
        return self.decode(frames)[:, :, :x.shape[-1]]

    def set_target_bandwidth(self, bandwidth: float):
        if bandwidth not in self.target_bandwidths:
            raise ValueError(f"This model doesn't support the bandwidth {bandwidth}. "
                             f"Select one of {self.target_bandwidths}.")
        self.bandwidth = bandwidth

    def get_lm_model(self) -> LMModel:
        """Return the associated LM model to improve the compression rate.
        """
        device = next(self.parameters()).device
        lm = LMModel(self.quantizer.n_q, self.quantizer.bins, num_layers=5, dim=200,
                     past_context=int(3.5 * self.frame_rate)).to(device)
        checkpoints = {
            'encodec_24khz': 'encodec_lm_24khz-1608e3c0.th',
            'encodec_48khz': 'encodec_lm_48khz-7add9fc3.th',
        }
        try:
            checkpoint_name = checkpoints[self.name]
        except KeyError:
            raise RuntimeError("No LM pre-trained for the current Encodec model.")
        url = _get_checkpoint_url(ROOT_URL, checkpoint_name)
        state = torch.hub.load_state_dict_from_url(
            url, map_location='cpu', check_hash=True)  # type: ignore
        lm.load_state_dict(state)
        lm.eval()
        return lm

    @staticmethod
    def _get_model(target_bandwidths: tp.List[float],
                   sample_rate: int = 24_000,
                   channels: int = 1,
                   causal: bool = True,
                   model_norm: str = 'weight_norm',
                   audio_normalize: bool = False,
                   segment: tp.Optional[float] = None,
                   name: str = 'unset'):
        encoder = m.SEANetEncoder(channels=channels, norm=model_norm, causal=causal)
        decoder = m.SEANetDecoder(channels=channels, norm=model_norm, causal=causal)
        n_q = int(1000 * target_bandwidths[-1] // (math.ceil(sample_rate / encoder.hop_length) * 10))
        quantizer = qt.ResidualVectorQuantizer(
            dimension=encoder.dimension,
            n_q=n_q,
            bins=1024,
        )
        model = EncodecModel(
            encoder,
            decoder,
            quantizer,
            target_bandwidths,
            sample_rate,
            channels,
            normalize=audio_normalize,
            segment=segment,
            name=name,
        )
        return model

    @staticmethod
    def _get_pretrained(checkpoint_name: str, repository: tp.Optional[Path] = None):
        if repository is not None:
            if not repository.is_dir():
                raise ValueError(f"{repository} must exist and be a directory.")
            file = repository / checkpoint_name
            checksum = file.stem.split('-')[1]
            _check_checksum(file, checksum)
            return torch.load(file)
        else:
            url = _get_checkpoint_url(ROOT_URL, checkpoint_name)
            return torch.hub.load_state_dict_from_url(url, map_location='cpu', check_hash=True)  # type:ignore

    @staticmethod
    def encodec_model_24khz(pretrained: bool = True, repository: tp.Optional[Path] = None):
        """Return the pretrained causal 24khz model.
        """
        if repository:
            assert pretrained
        target_bandwidths = [1.5, 3., 6, 12., 24.]
        checkpoint_name = 'encodec_24khz-d7cc33bc.th'
        sample_rate = 24_000
        channels = 1
        model = EncodecModel._get_model(
            target_bandwidths, sample_rate, channels,
            causal=True, model_norm='weight_norm', audio_normalize=False,
            name='encodec_24khz' if pretrained else 'unset')
        if pretrained:
            state_dict = EncodecModel._get_pretrained(checkpoint_name, repository)
            model.load_state_dict(state_dict)
        model.eval()
        return model

    @staticmethod
    def encodec_model_48khz(pretrained: bool = True, repository: tp.Optional[Path] = None):
        """Return the pretrained 48khz model.
        """
        if repository:
            assert pretrained
        target_bandwidths = [3., 6., 12., 24.]
        checkpoint_name = 'encodec_48khz-7e698e3e.th'
        sample_rate = 48_000
        channels = 2
        model = EncodecModel._get_model(
            target_bandwidths, sample_rate, channels,
            causal=False, model_norm='time_group_norm', audio_normalize=True,
            segment=1., name='encodec_48khz' if pretrained else 'unset')
        if pretrained:
            state_dict = EncodecModel._get_pretrained(checkpoint_name, repository)
            model.load_state_dict(state_dict)
        model.eval()
        return model


def test():
    from itertools import product
    import torchaudio
    bandwidths = [3, 6, 12, 24]
    models = {
        'encodec_24khz': EncodecModel.encodec_model_24khz,
        'encodec_48khz': EncodecModel.encodec_model_48khz
    }
    for model_name, bw in product(models.keys(), bandwidths):
        model = models[model_name]()
        model.set_target_bandwidth(bw)
        audio_suffix = model_name.split('_')[1][:3]
        wav, sr = torchaudio.load(f"test_{audio_suffix}.wav")
        wav = wav[:, :model.sample_rate * 2]
        wav_in = wav.unsqueeze(0)
        wav_dec = model(wav_in)[0]
        assert wav.shape == wav_dec.shape, (wav.shape, wav_dec.shape)


if __name__ == '__main__':
    test()



#!/usr/bin/env python3
# Copyright (c) 2024 Alibaba Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import torch
import torchaudio
from tqdm import tqdm
import onnxruntime
import torchaudio.compliance.kaldi as kaldi


def main(args):
    utt2wav, utt2spk = {}, {}
    with open('{}/wav.scp'.format(args.dir)) as f:
        for l in f:
            l = l.replace('\n', '').split()
            utt2wav[l[0]] = l[1]
    with open('{}/utt2spk'.format(args.dir)) as f:
        for l in f:
            l = l.replace('\n', '').split()
            utt2spk[l[0]] = l[1]

    option = onnxruntime.SessionOptions()
    option.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
    option.intra_op_num_threads = 1
    providers = ["CPUExecutionProvider"]
    ort_session = onnxruntime.InferenceSession(args.onnx_path, sess_options=option, providers=providers)

    utt2embedding, spk2embedding = {}, {}
    for utt in tqdm(utt2wav.keys()):
        audio, sample_rate = torchaudio.load(utt2wav[utt])
        if sample_rate != 16000:
            audio = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=16000)(audio)
        feat = kaldi.fbank(audio,
                           num_mel_bins=80,
                           dither=0,
                           sample_frequency=16000)
        feat = feat - feat.mean(dim=0, keepdim=True)
        embedding = ort_session.run(None, {ort_session.get_inputs()[0].name: feat.unsqueeze(dim=0).cpu().numpy()})[0].flatten().tolist()
        utt2embedding[utt] = embedding
        spk = utt2spk[utt]
        if spk not in spk2embedding:
            spk2embedding[spk] = []
        spk2embedding[spk].append(embedding)
    for k, v in spk2embedding.items():
        spk2embedding[k] = torch.tensor(v).mean(dim=0).tolist()

    torch.save(utt2embedding, '{}/utt2embedding.pt'.format(args.dir))
    torch.save(spk2embedding, '{}/spk2embedding.pt'.format(args.dir))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--dir',
                        type=str)
    parser.add_argument('--onnx_path',
                        type=str)
    args = parser.parse_args()
    main(args)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
import re
from dataclasses import dataclass, field
from typing import List, Optional

import torch
import torch.nn.functional as F
from fairseq import metrics, utils
from fairseq.models.hubert import ILSHubertModel
from fairseq.criterions import FairseqCriterion, register_criterion
from fairseq.dataclass import FairseqDataclass


@dataclass
class HubertCriterionConfig(FairseqDataclass):
    pred_masked_weight: float = field(
        default=1.0,
        metadata={"help": "weight for predictive loss for masked frames"},
    )
    pred_nomask_weight: float = field(
        default=0.0,
        metadata={"help": "weight for predictive loss for unmasked frames"},
    )
    loss_weights: Optional[List[float]] = field(
        default=None,
        metadata={"help": "weights for additional loss terms (not first one)"},
    )
    log_keys: List[str] = field(
        default_factory=lambda: [],
        metadata={"help": "output keys to log"},
    )


@register_criterion("hubert", dataclass=HubertCriterionConfig)
class HubertCriterion(FairseqCriterion):
    def __init__(self, task, pred_masked_weight, pred_nomask_weight, loss_weights=None, log_keys=None):
        super().__init__(task)
        self.pred_masked_weight = pred_masked_weight
        self.pred_nomask_weight = pred_nomask_weight
        self.loss_weights = loss_weights
        self.log_keys = [] if log_keys is None else log_keys

    def get_net_output(self, model, sample):
        """compute the loss for the given sample"""
        net_output = model(target_list=sample["target_list"], **sample["net_input"])
        return net_output

    def get_loss(self, model, sample, net_output, reduce=True, log_pred=False):
        """Compute the loss for the given sample.
        Returns a tuple with three elements:
        1) the loss
        2) the sample size, which is used as the denominator for the gradient
        3) logging outputs to display while training
        """
        loss = 0.
        sample_size = 0
        logging_output = {}
        reduction = "sum" if reduce else "none"

        loss_m_list = []
        logp_m_list = model.get_logits(net_output, True)
        targ_m_list = model.get_targets(net_output, True)
        assert self.pred_masked_weight == 0 or len(logp_m_list) > 0
        for i, (logp_m, targ_m) in enumerate(zip(logp_m_list, targ_m_list)):
            loss_m = F.cross_entropy(logp_m, targ_m, reduction=reduction)
            loss_m_list.append(loss_m)
            logging_output[f"loss_m_{i}"] = loss_m.detach().item()
        if self.pred_masked_weight > 0:
            if isinstance(model, ILSHubertModel):
                if model.weighted_sum:
                    norm_weights = F.softmax(model.weights, dim=-1)
                    loss_m_list = norm_weights * torch.stack(loss_m_list, dim=0)
            loss += self.pred_masked_weight * sum(loss_m_list)
            sample_size += targ_m_list[0].numel()

        loss_u_list = []
        logp_u_list = model.get_logits(net_output, False)
        targ_u_list = model.get_targets(net_output, False)
        assert self.pred_nomask_weight == 0 or len(logp_u_list) > 0
        for i, (logp_u, targ_u) in enumerate(zip(logp_u_list, targ_u_list)):
            loss_u = F.cross_entropy(logp_u, targ_u, reduction=reduction)
            loss_u_list.append(loss_u)
            logging_output[f"loss_u_{i}"] = loss_u.detach().item()
        if self.pred_nomask_weight > 0:
            if model.weighted_sum:
                norm_weights = F.softmax(model.weights, dim=-1)
                loss_u_list = norm_weights * torch.stack(loss_u_list, dim=0)
            loss += self.pred_nomask_weight * sum(loss_u_list)
            sample_size += targ_u_list[0].numel()

        if self.loss_weights is not None:
            assert hasattr(model, "get_extra_losses")
            extra_losses, names = model.get_extra_losses(net_output)
            if torch.is_tensor(extra_losses):
                extra_losses = [extra_losses]
                names = [names]
            if len(self.loss_weights) == 1 and len(extra_losses) != 1:
                self.loss_weights = [self.loss_weights[0]] * len(extra_losses)
            assert len(extra_losses) == len(self.loss_weights), f"{len(extra_losses)}, {len(self.loss_weights)}"
            for p, n, coef in zip(extra_losses, names, self.loss_weights):
                if coef != 0 and p is not None:
                    p = coef * p.float() * sample_size
                    loss += p
                    logging_output[f"loss_{n}"] = p.item()

        logging_output = {
            "loss": loss.item() if reduce else loss,
            "ntokens": sample_size,
            "nsentences": sample["id"].numel(),
            "sample_size": sample_size,
            **logging_output,
        }

        for lk in self.log_keys:
            if lk in net_output:
                logging_output[lk] = float((net_output[lk]))

        def compute_correct(logits):
            if logits.numel() == 0:
                return 0, 0
            else:
                assert logits.dim() > 1, logits.shape
                max = logits.argmax(-1) == 0
                min = logits.argmin(-1) == 0
                both = max & min
                corr = max.long().sum().item() - both.long().sum().item()
                count = max.numel()
                return corr, count

        with torch.no_grad():
            for i, logp_m in enumerate(logp_m_list):
                corr_m, count_m = compute_correct(logp_m)
                logging_output[f"correct_m_{i}"] = corr_m
                logging_output[f"count_m_{i}"] = count_m

            for i, logp_u in enumerate(logp_u_list):
                corr_u, count_u = compute_correct(logp_u)
                logging_output[f"correct_u_{i}"] = corr_u
                logging_output[f"count_u_{i}"] = count_u

        return loss, sample_size, logging_output

    def forward(self, model, sample, reduce=True, log_pred=False):
        net_output = self.get_net_output(model, sample)
        loss, sample_size, logging_output = self.get_loss(model, sample, net_output, reduce, log_pred)
        return loss, sample_size, logging_output
        

    @staticmethod
    def reduce_metrics(logging_outputs) -> None:
        """Aggregate logging outputs from data parallel training (copied from normal cross entropy)."""
        loss_sum = sum(log.get("loss", 0) for log in logging_outputs)
        ntokens = sum(log.get("ntokens", 0) for log in logging_outputs)
        sample_size = sum(log.get("sample_size", 0) for log in logging_outputs)

        metrics.log_scalar("loss", loss_sum / sample_size / math.log(2), sample_size, round=3)
        if sample_size != ntokens:
            metrics.log_scalar("nll_loss", loss_sum / ntokens / math.log(2), ntokens, round=3)
            metrics.log_derived("ppl", lambda meters: utils.get_perplexity(meters["nll_loss"].avg))
        else:
            metrics.log_derived("ppl", lambda meters: utils.get_perplexity(meters["loss"].avg))

        counts = {}
        for lk in logging_outputs[0].keys():
            if lk.startswith("count_"):
                val = sum(log[lk] for log in logging_outputs)
                metrics.log_scalar(lk, val)
                counts[lk] = val

        for lk in logging_outputs[0].keys():
            if lk.startswith("loss_"):
                val = sum(log[lk] for log in logging_outputs)
                metrics.log_scalar(lk, val / sample_size / math.log(2), round=3)
            elif lk.startswith("correct_"):
                val = sum(log[lk] for log in logging_outputs)
                metrics.log_scalar(lk, val / counts[re.sub("correct", "count", lk)])

    @staticmethod
    def aggregate_logging_outputs(logging_outputs):
        """Aggregate logging outputs from data parallel training."""
        raise NotImplementedError()

    @staticmethod
    def logging_outputs_can_be_summed() -> bool:
        """
        Whether the logging outputs returned by `forward` can be summed
        across workers prior to calling `reduce_metrics`. Setting this
        to True will improves distributed training speed.
        """
        return False



"""
A model worker with transformers libs executes the model.

Run BF16 inference with:

python model_server.py --host localhost --model-path THUDM/glm-4-voice-9b --port 10000 --dtype bfloat16 --device cuda:0

Run Int4 inference with:

python model_server.py --host localhost --model-path THUDM/glm-4-voice-9b --port 10000 --dtype int4 --device cuda:0

"""
import argparse
import json

from fastapi import FastAPI, Request
from fastapi.responses import StreamingResponse
from transformers import AutoModel, AutoTokenizer, BitsAndBytesConfig
from transformers.generation.streamers import BaseStreamer
import torch
import uvicorn

from threading import Thread
from queue import Queue


class TokenStreamer(BaseStreamer):
    def __init__(self, skip_prompt: bool = False, timeout=None):
        self.skip_prompt = skip_prompt

        # variables used in the streaming process
        self.token_queue = Queue()
        self.stop_signal = None
        self.next_tokens_are_prompt = True
        self.timeout = timeout

    def put(self, value):
        if len(value.shape) > 1 and value.shape[0] > 1:
            raise ValueError("TextStreamer only supports batch size 1")
        elif len(value.shape) > 1:
            value = value[0]

        if self.skip_prompt and self.next_tokens_are_prompt:
            self.next_tokens_are_prompt = False
            return

        for token in value.tolist():
            self.token_queue.put(token)

    def end(self):
        self.token_queue.put(self.stop_signal)

    def __iter__(self):
        return self

    def __next__(self):
        value = self.token_queue.get(timeout=self.timeout)
        if value == self.stop_signal:
            raise StopIteration()
        else:
            return value


class ModelWorker:
    def __init__(self, model_path, dtype="bfloat16", device='cuda'):
        self.device = device
        self.bnb_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_use_double_quant=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_compute_dtype=torch.bfloat16
        ) if dtype == "int4" else None

        self.glm_model = AutoModel.from_pretrained(
            model_path,
            trust_remote_code=True,
            quantization_config=self.bnb_config if self.bnb_config else None,
            device_map={"": 0}
        ).eval()
        self.glm_tokenizer = AutoTokenizer.from_pretrained(model_path, trust_remote_code=True)

    @torch.inference_mode()
    def generate_stream(self, params):
        tokenizer, model = self.glm_tokenizer, self.glm_model

        prompt = params["prompt"]

        temperature = float(params.get("temperature", 1.0))
        top_p = float(params.get("top_p", 1.0))
        max_new_tokens = int(params.get("max_new_tokens", 256))

        inputs = tokenizer([prompt], return_tensors="pt")
        inputs = inputs.to(self.device)
        streamer = TokenStreamer(skip_prompt=True)
        thread = Thread(
            target=model.generate,
            kwargs=dict(
                **inputs,
                max_new_tokens=int(max_new_tokens),
                temperature=float(temperature),
                top_p=float(top_p),
                streamer=streamer
            )
        )
        thread.start()
        for token_id in streamer:
            yield (json.dumps({"token_id": token_id, "error_code": 0}) + "\n").encode()

    def generate_stream_gate(self, params):
        try:
            for x in self.generate_stream(params):
                yield x
        except Exception as e:
            print("Caught Unknown Error", e)
            ret = {
                "text": "Server Error",
                "error_code": 1,
            }
            yield (json.dumps(ret) + "\n").encode()


app = FastAPI()


@app.post("/generate_stream")
async def generate_stream(request: Request):
    params = await request.json()

    generator = worker.generate_stream_gate(params)
    return StreamingResponse(generator)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--dtype", type=str, default="bfloat16")
    parser.add_argument("--device", type=str, default="cuda:0")
    parser.add_argument("--port", type=int, default=10000)
    parser.add_argument("--model-path", type=str, default="THUDM/glm-4-voice-9b")
    args = parser.parse_args()

    worker = ModelWorker(args.model_path, args.dtype, args.device)
    uvicorn.run(app, host=args.host, port=args.port, log_level="info")



#
# Copyright (c) 2024–2025, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

import json
from typing import Any, Dict, List


class FunctionSchema:
    """Standardized function schema representation for tool definition.

    Provides a structured way to define function tools used with AI models like OpenAI.
    This schema defines the function's name, description, parameter properties, and
    required parameters, following specifications required by AI service providers.

    Args:
        name: Name of the function to be called.
        description: Description of what the function does.
        properties: Dictionary defining parameter types, descriptions, and constraints.
        required: List of property names that are required parameters.
    """

    def __init__(
        self, name: str, description: str, properties: Dict[str, Any], required: List[str]
    ) -> None:
        self._name = name
        self._description = description
        self._properties = properties
        self._required = required

    def __str__(self):
        return json.dumps(self.to_default_dict())

    def to_default_dict(self) -> Dict[str, Any]:
        """Converts the function schema to a dictionary.

        Returns:
            Dictionary representation of the function schema.
        """
        return {
            "name": self._name,
            "description": self._description,
            "parameters": {
                "type": "object",
                "properties": self._properties,
                "required": self._required,
            },
        }

    @property
    def name(self) -> str:
        """Get the function name.

        Returns:
            The function name.
        """
        return self._name

    @property
    def description(self) -> str:
        """Get the function description.

        Returns:
            The function description.
        """
        return self._description

    @property
    def properties(self) -> Dict[str, Any]:
        """Get the function properties.

        Returns:
            Dictionary of parameter specifications.
        """
        return self._properties

    @property
    def required(self) -> List[str]:
        """Get the required parameters.

        Returns:
            List of required parameter names.
        """
        return self._required



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""isort:skip_file"""

import importlib
import os

from fairseq import registry
from fairseq.optim.bmuf import FairseqBMUF  # noqa
from fairseq.optim.fairseq_optimizer import (  # noqa
    FairseqOptimizer,
    LegacyFairseqOptimizer,
)
from fairseq.optim.amp_optimizer import AMPOptimizer
from fairseq.optim.fp16_optimizer import FP16Optimizer, MemoryEfficientFP16Optimizer
from fairseq.optim.shard import shard_
from omegaconf import DictConfig

__all__ = [
    "AMPOptimizer",
    "FairseqOptimizer",
    "FP16Optimizer",
    "MemoryEfficientFP16Optimizer",
    "shard_",
]

(
    _build_optimizer,
    register_optimizer,
    OPTIMIZER_REGISTRY,
    OPTIMIZER_DATACLASS_REGISTRY,
) = registry.setup_registry("--optimizer", base_class=FairseqOptimizer, required=True)


def build_optimizer(cfg: DictConfig, params, *extra_args, **extra_kwargs):
    if all(isinstance(p, dict) for p in params):
        params = [t for p in params for t in p.values()]
    params = list(filter(lambda p: p.requires_grad, params))
    return _build_optimizer(cfg, params, *extra_args, **extra_kwargs)


# automatically import any Python files in the optim/ directory
for file in sorted(os.listdir(os.path.dirname(__file__))):
    if file.endswith(".py") and not file.startswith("_"):
        file_name = file[: file.find(".py")]
        importlib.import_module("fairseq.optim." + file_name)



import traceback
import logging
import threading
import queue

import grpc

from src.common.grpc.idl.connector_pb2 import ConnectStreamRequest
from src.common.grpc.idl.connector_pb2_grpc import ConnectorStub
from src.common.grpc.interceptors.authentication_client import add_authentication


class StreamClient:
    def __init__(
        self,
        in_q: queue.Queue,
        out_q: queue.Queue,
        token="chat-bot-connenctor",
        target="localhost:50052",
    ) -> None:
        self.in_q = in_q
        self.out_q = out_q
        self.token = token
        self.target = target
        self.connect_thread = None

    def start(self):
        if self.connect_thread is None or self.connect_thread.is_alive() is False:
            self.connect_thread = threading.Thread(target=self._connect, args=())
            self.connect_thread.start()

    def stop(self):
        if self.connect_thread is not None and self.connect_thread.is_alive():
            self.connect_thread.join()
            logging.info("stream client thread stop")

    def _connect(self):
        authentication = add_authentication("authorization", self.token)
        channel = grpc.insecure_channel(self.target)
        channel = grpc.intercept_channel(channel, authentication)
        stub = ConnectorStub(channel)

        def iterator():
            while True:
                frame = self.in_q.get()
                if frame is None:
                    logging.debug("request frame is None, break")
                    break
                logging.debug(f"fe in_q get <---- len(frame):{len(frame)}")
                yield ConnectStreamRequest(frame=frame)

        try:
            response_iter = stub.ConnectStream(iterator())
            for response in response_iter:
                if response.frame is None:
                    logging.debug("response frame is None, break")
                    break
                logging.debug(f"fe out_q put ----> len(response.frame):{len(response.frame)}")
                self.out_q.put(response.frame)
        except Exception as e:
            logging.error(e)
            logging.error(traceback.format_exc())

    def close(self):
        self.in_q.put(None)
        self.stop()
        logging.debug("stream client stoped")



import logging
from queue import Queue, Empty
from typing import Generator, AsyncGenerator

from src.common.factory import EngineClass
from src.common.types import AudioRecoderArgs, PYAUDIO_PACONTINUE
from src.common.interface import IAudioStream
from src.types.speech.audio_stream import AudioStreamInfo


class AudioRecorder(EngineClass):
    buffer_queue = Queue()

    @classmethod
    def stream_callback(cls, in_data, frame_count=0, time_info=None, status=None):
        # print(len(in_data), frame_count, time_info, status, cls.buffer_queue and cls.buffer_queue.qsize())
        cls.buffer_queue and cls.buffer_queue.put_nowait(in_data)
        play_data = chr(0) * len(in_data)
        return play_data, PYAUDIO_PACONTINUE

    def __init__(self, **args) -> None:
        self.args = AudioRecoderArgs(**args)

    def get_record_buf(self) -> bytes:
        if self.args.is_stream_callback is False:
            return self.audio.read_stream(self.args.num_frames)
        # !NOTE: no block for audio check is_stream_active
        return self.buffer_queue.get(timeout=0.1)

    def frame_genrator(self) -> Generator[bytes, None, None]:
        while self.audio.is_stream_active():
            try:
                frames = self.get_record_buf()
                yield frames
            except Empty:
                continue
        logging.info("frame_genrator end")
        # yield None

    def open(self):
        self.audio.open_stream()

    def close(self):
        self.audio.close()

    def set_in_stream(self, audio_stream: EngineClass | IAudioStream):
        stream_callback = AudioRecorder.stream_callback if self.args.is_stream_callback else None
        audio_stream.set_args(
            stream_callback=stream_callback,
        )
        self.audio = audio_stream
        self.stream_info: AudioStreamInfo = self.audio.get_stream_info()



import os
import subprocess
from threading import Thread
from time import perf_counter


import modal


app = modal.App("fastvlm")
img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "clang", "cmake", "ninja-build")
    .run_commands(
        "git clone -b feat/achatbot https://github.com/weedge/ml-fastvlm.git",
        "cd /ml-fastvlm && git checkout 6012455cb7723d6be90cb9eaf462325e6d4a1849",
        "cd /ml-fastvlm && pip install -e .",
    )
    .pip_install(
        "torch==2.6.0",
        "torchaudio==2.6.0",
        "torchvision==0.21.0",
        index_url="https://download.pytorch.org/whl/cu126",
    )
    .pip_install("wheel")
    # https://github.com/Dao-AILab/flash-attention/releases/tag/v2.7.4.post1
    .pip_install("flash-attn", extra_options="--no-build-isolation")
    .env(
        {
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
            "TQDM_DISABLE": "1",
            "LLM_MODEL": os.getenv("LLM_MODEL", "llava-fastvithd_1.5b_stage3"),
            "MOBILE_CLIP_MODEL_CONFIG": "/root/.achatbot/models/mobileclip_l.json",
        }
    )
)

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)


with img.imports():
    import torch
    from PIL import Image
    from transformers.generation.streamers import BaseStreamer, TextIteratorStreamer

    from llava.utils import disable_torch_init
    from llava.conversation import conv_templates
    from llava.model.builder import load_pretrained_model
    from llava.mm_utils import tokenizer_image_token, process_images, get_model_name_from_path
    from llava.constants import (
        IMAGE_TOKEN_INDEX,
        DEFAULT_IMAGE_TOKEN,
        DEFAULT_IM_START_TOKEN,
        DEFAULT_IM_END_TOKEN,
    )

    MODEL_PATH = os.path.join(HF_MODEL_DIR, os.getenv("LLM_MODEL"))


@app.function(
    gpu=os.getenv("IMAGE_GPU", None),
    cpu=2.0,
    retries=1,
    image=img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        ASSETS_DIR: assets_dir,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
def run(func):
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    gpu_prop = torch.cuda.get_device_properties("cuda")
    print(gpu_prop)

    func()


def print_model_params(model: torch.nn.Module, extra_info=""):
    # print the number of parameters in the model
    model_million_params = sum(p.numel() for p in model.parameters()) / 1e6
    print(model)
    print(f"{extra_info} {model_million_params} M parameters")


def dump_model():
    for model_name in [
        "llava-fastvithd_0.5b_stage3",
        "llava-fastvithd_1.5b_stage3",
        "llava-fastvithd_7b_stage3",
    ]:
        model_path = os.path.join(HF_MODEL_DIR, model_name)
        # Load model
        disable_torch_init()
        model_name = get_model_name_from_path(model_path)
        print(model_name)
        device = "mps" if torch.backends.mps.is_available() else "cuda"
        tokenizer, model, image_processor, context_len = load_pretrained_model(
            model_path, None, model_name, device=device
        )
        model = model.eval()
        print(f"{model.config=}")
        print(f"{tokenizer=}")
        print_model_params(model, f"LlavaQwen2ForCausalLM_{model_name}")

        del model
        torch.cuda.empty_cache()


# https://github.com/apple/ml-fastvit
# https://github.com/apple/ml-mobileclip
# https://github.com/apple/ml-fastvlm
def predict():
    # Load model
    disable_torch_init()
    model_name = get_model_name_from_path(MODEL_PATH)
    print(f"{model_name=}")
    device = "mps" if torch.backends.mps.is_available() else "cuda"
    print(f"device {device}")
    tokenizer, model, image_processor, context_len = load_pretrained_model(
        MODEL_PATH, None, model_name, device=device
    )
    model = model.eval()
    print(f"model device {model.device}")
    print(f"{model.config=}")
    print(f"{tokenizer=}")
    # print(f"{model=}")

    # Construct prompt
    qs = "请用中文描述图片内容"
    qs = "Please reply to my message in Chinese simplified(简体中文), don't use Markdown format. 描述下图片内容"
    if model.config.mm_use_im_start_end:
        qs = DEFAULT_IM_START_TOKEN + DEFAULT_IMAGE_TOKEN + DEFAULT_IM_END_TOKEN + "\n" + qs
    else:
        qs = DEFAULT_IMAGE_TOKEN + "\n" + qs
    conv_mode = "qwen_2"
    conv = conv_templates[conv_mode].copy()
    conv.append_message(conv.roles[0], qs)
    conv.append_message(conv.roles[1], None)
    prompt = conv.get_prompt()
    print(f"{prompt=}")

    # Set the pad token id for generation
    model.generation_config.pad_token_id = tokenizer.pad_token_id

    # Tokenize prompt
    input_ids = (
        tokenizer_image_token(prompt, tokenizer, IMAGE_TOKEN_INDEX, return_tensors="pt")
        .unsqueeze(0)
        .to(torch.device(device))
    )

    image_file = "/ml-fastvlm/docs/acc_vs_latency_qwen-2.png"
    # Load and preprocess image
    image = Image.open(image_file).convert("RGB")
    image_tensor = process_images([image], image_processor, model.config)[0]

    # Run inference
    with torch.inference_mode():
        output_ids = model.generate(
            input_ids,
            images=image_tensor.unsqueeze(0).half(),
            image_sizes=[image.size],
            do_sample=True,
            temperature=0.2,
            top_p=None,
            num_beams=1,
            max_new_tokens=1024,
            use_cache=True,
        )

        outputs = tokenizer.batch_decode(output_ids, skip_special_tokens=True)[0].strip()
        print(outputs)


# https://github.com/apple/ml-fastvit
# https://github.com/apple/ml-mobileclip
# https://github.com/apple/ml-fastvlm
def predict_stream():
    # Load model
    disable_torch_init()
    model_name = get_model_name_from_path(MODEL_PATH)
    print(model_name)
    device = "mps" if torch.backends.mps.is_available() else "cuda"
    tokenizer, model, image_processor, context_len = load_pretrained_model(
        MODEL_PATH, None, model_name, device=device
    )
    model = model.eval()
    print(f"{model.config=}")
    print(f"{tokenizer=}")
    print(f"{model=}")

    # Construct prompt
    qs = "Please reply to my message in Chinese simplified(简体中文), Do not use Markdown format. 描述下图片内容"
    if model.config.mm_use_im_start_end:
        qs = DEFAULT_IM_START_TOKEN + DEFAULT_IMAGE_TOKEN + DEFAULT_IM_END_TOKEN + "\n" + qs
    else:
        qs = DEFAULT_IMAGE_TOKEN + "\n" + qs
    conv_mode = "qwen_2"
    conv = conv_templates[conv_mode].copy()
    conv.append_message(conv.roles[0], qs)
    conv.append_message(conv.roles[1], None)
    prompt = conv.get_prompt()
    print(f"{prompt=}")
    print(f"{conv.messages=}")

    # Set the pad token id for generation
    model.generation_config.pad_token_id = tokenizer.pad_token_id

    # Tokenize prompt
    input_ids = (
        tokenizer_image_token(prompt, tokenizer, IMAGE_TOKEN_INDEX, return_tensors="pt")
        .unsqueeze(0)
        .to(torch.device(device))
    )
    print(input_ids)

    image_file = "/ml-fastvlm/docs/acc_vs_latency_qwen-2.png"
    # Load and preprocess image
    image = Image.open(image_file).convert("RGB")
    image_tensor = process_images([image], image_processor, model.config)
    print(f"{image_tensor.shape=}")

    for i in range(2):
        streamer = TextIteratorStreamer(
            tokenizer=tokenizer, skip_prompt=True, skip_special_tokens=True
        )
        generation_kwargs = dict(
            inputs=input_ids,
            images=image_tensor.half().to(torch.device(device)),
            image_sizes=[image.size],
            do_sample=True,
            temperature=0.2,
            top_p=None,
            num_beams=1,
            repetition_penalty=1.5,
            max_new_tokens=1024,
            use_cache=True,
            streamer=streamer,
        )
        thread = Thread(target=model.generate, kwargs=generation_kwargs)
        thread.start()
        generated_text = ""
        start = perf_counter()
        times = []
        for new_text in streamer:
            times.append(perf_counter() - start)
            print(new_text, end="", flush=True)
            generated_text += new_text
            start = perf_counter()
        print(f"\n{i}. {generated_text=} TTFT: {times[0]:.2f}s total time: {sum(times):.2f}s")


"""
IMAGE_GPU=T4 modal run src/llm/transformers/vlm/fastvlm.py --task dump_model
IMAGE_GPU=T4 modal run src/llm/transformers/vlm/fastvlm.py --task predict
"""


@app.local_entrypoint()
def main(task: str = "dump_model"):
    tasks = {
        "dump_model": dump_model,
        "predict": predict,
        "predict_stream": predict_stream,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    run.remote(tasks[task])



import logging
import random
import os
import sys
from typing import AsyncGenerator
from importlib.resources import files

from dotenv import load_dotenv
import numpy as np
import torch
import torchaudio

from src.common.utils import audio_utils
from src.common.interface import ITts
from src.common.session import Session
from src.common.types import ASSETS_DIR, PYAUDIO_PAINT16, PYAUDIO_PAFLOAT32
from src.types.speech.tts.f5 import F5TTSArgs
from .base import BaseTTS

load_dotenv(override=True)

try:
    cur_dir = os.path.dirname(__file__)
    if bool(os.getenv("ACHATBOT_PKG", "")):
        sys.path.insert(1, os.path.join(cur_dir, "../../../F5TTS/src"))
    else:
        sys.path.insert(1, os.path.join(cur_dir, "../../../../deps/F5TTS/src"))

    from deps.F5TTS.src.f5_tts.infer.utils_infer import (
        load_vocoder,
        load_model,
        preprocess_ref_audio_text,
        infer_process,
        infer_batch_process,
    )
    from deps.F5TTS.src.f5_tts.model.backbones.dit import DiT
    from deps.F5TTS.src.f5_tts.model.backbones.unett import UNetT
    from deps.F5TTS.src.f5_tts.model.utils import seed_everything
except ModuleNotFoundError as e:
    logging.error("In order to use f5-tts, you need to `pip install achatbot[tts_f5]`.")
    raise Exception(f"Missing module: {e}")


class F5TTS(BaseTTS, ITts):
    r"""
    https://github.com/ai-bot-pro/achatbot/pull/101
    """

    TAG = "tts_f5"

    @classmethod
    def get_args(cls, **kwargs) -> dict:
        return {**F5TTSArgs().__dict__, **kwargs}

    def __init__(self, **args) -> None:
        self.args = F5TTSArgs(**args)
        self.args.device = self.args.device or (
            "cuda"
            if torch.cuda.is_available()
            else "mps"
            if torch.backends.mps.is_available()
            else "cpu"
        )

        # Load the vocoder model
        self.vocoder_model = load_vocoder(
            self.args.vocoder_name,
            self.args.vocoder_ckpt_dir is not None,
            self.args.vocoder_ckpt_dir,
            self.args.device,
        )
        # print the number of parameters in the model
        model_million_params = sum(p.numel() for p in self.vocoder_model.parameters()) / 1e6
        logging.debug(
            f"{F5TTS.TAG} vocoder_model params: {model_million_params} M {self.vocoder_model}"
        )

        if self.args.model_type == "E2-TTS":
            model_cls = UNetT
        else:
            model_cls = DiT

        if self.args.vocab_file == "":
            self.args.vocab_file = os.path.join(ASSETS_DIR, "vocab.txt")

        # Load the CFM model using the provided checkpoint and vocab files
        self.cfm_model = load_model(
            model_cls=model_cls,
            model_cfg=self.args.model_cfg,
            ckpt_path=self.args.model_ckpt_path,
            mel_spec_type=self.args.vocoder_name,
            vocab_file=self.args.vocab_file,
            ode_method=self.args.ode_method,
            use_ema=self.args.use_EMA,
            device=self.args.device,
        )
        # print the number of parameters in the model
        model_million_params = sum(p.numel() for p in self.cfm_model.parameters()) / 1e6
        logging.debug(f"{F5TTS.TAG} CFM model params: {model_million_params} M {self.cfm_model}")

        # torch manual seed
        if self.args.seed == -1:
            self.args.seed = random.randint(0, 4294967295)
        seed_everything(self.args.seed)

        # inference ref
        if self.args.ref_audio_file == "":
            self.args.ref_audio_file = os.path.join(ASSETS_DIR, "basic_ref_en.wav")
            self.args.ref_text = "some call me nature, others call me mother nature."

        if self.args.preprocess_ref_audio_text is True:
            self.args.ref_file, self.args.ref_text = preprocess_ref_audio_text(
                self.args.ref_audio_file, self.args.ref_text, device=self.args.device
            )
        logging.debug(f"{F5TTS.TAG} args: {self.args}")

    def _warm_up(self):
        """
        Warm up the model with a dummy input to ensure it's ready for real-time processing.
        """
        logging.info("Warming up the f5-tts model...")
        audio, sr = torchaudio.load(self.args.ref_audio_file)
        gen_text = "Warm-up text for the model."

        # Pass the vocoder as an argument here
        infer_batch_process(
            (audio, sr),
            self.args.ref_text,
            [gen_text],
            self.cfm_model,
            self.vocoder_model,
            device=self.args.device,
        )
        logging.info("Warm-up completed.")

    def get_stream_info(self) -> dict:
        return {
            # "format": PYAUDIO_PAINT16,
            "format": PYAUDIO_PAFLOAT32,
            "channels": 1,
            "rate": 24000,  # target_sample_rate
            "sample_width": 2,
            # "np_dtype": np.int16,
            "np_dtype": np.float32,
        }

    def set_voice(self, ref_file: str):
        # use ASR transcribe ref audio to text
        self.args.ref_file, self.args.ref_text = preprocess_ref_audio_text(
            ref_file, "", device=self.args.device
        )

    def get_voices(self):
        return [self.args.ref_text]

    @torch.inference_mode()
    async def _inference(
        self, session: Session, text: str, **kwargs
    ) -> AsyncGenerator[bytes, None]:
        wav, _, _ = infer_process(
            self.args.ref_audio_file,
            self.args.ref_text,
            text,
            self.cfm_model,
            self.vocoder_model,
            self.args.vocoder_name,
            target_rms=self.args.target_rms,
            cross_fade_duration=self.args.cross_fade_duration,
            nfe_step=self.args.nfe_step,
            cfg_strength=self.args.cfg_strength,
            sway_sampling_coef=self.args.sway_sampling_coef,
            speed=self.args.speed,
            fix_duration=self.args.fix_duration,
            device=self.args.device,
        )

        # yield audio_utils.postprocess_tts_wave_int16(torch.from_numpy(wav))

        # use float32 have high-quality wave
        # yield audio_utils.postprocess_tts_wave(torch.from_numpy(wav))
        yield np.frombuffer(wav, dtype=np.float32).tobytes()
